# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import datetime
from typing import cast

import numpy as np
import pytest

from otx.api.configuration import ConfigurableParameters
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelSchemaEntity
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    ColorPalette,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    Performance,
    ScoreMetric,
    TextChartInfo,
    TextMetricsGroup,
    VisualizationType,
)
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.evaluation.f_measure import (
    FMeasure,
    _AggregatedResults,
    _FMeasureCalculator,
    _Metrics,
    _OverallResults,
    _ResultCounters,
    bounding_box_intersection_over_union,
    get_iou_matrix,
    get_n_false_negatives,
    intersection_box,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestFMeasureFunctions:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_intersection_box(self):
        """
        <b>Description:</b>
        Check "intersection_box" function

        <b>Input data:</b>
        Bounding boxes coordinates

        <b>Expected results:</b>
        Test passes if box returned by "intersection_box" function has expected coordinates

        <b>Steps</b>
        1. Check box returned by "intersection_box" function for boxes that intersect in two points
        2. Check box returned by "intersection_box" function for boxes that intersect in one point
        3. Check box returned by "intersection_box" function for boxes that intersect by one side
        4. Check box returned by "intersection_box" function when one of boxes completely covers other
        """
        base_box = [2, 2, 5, 6]
        # Checking box returned by "intersection_box" for boxes that intersect in two points
        assert intersection_box(box1=base_box, box2=[4, 4, 7, 8]) == (4, 5, 6, 4)
        # Checking box returned by "intersection_box" for boxes that intersect in one point
        assert intersection_box(box1=base_box, box2=[1, 1, 2, 2]) == (2, 2, 2, 2)
        # Checking box returned by "intersection_box" for boxes that intersect by one side
        assert intersection_box(box1=base_box, box2=[2, 1, 5, 2]) == (2, 5, 2, 2)
        # Checking box returned by "intersection_box" when one of boxes completely covers other
        assert intersection_box(box1=base_box, box2=[0, 0, 10, 10]) == (2, 5, 6, 2)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_bounding_box_intersection_over_union(self):
        """
        <b>Description:</b>
        Check "bounding_box_intersection_over_union" function

        <b>Input data:</b>
        Bounding boxes coordinates

        <b>Expected results:</b>
        Test passes if value returned by "bounding_box_intersection_over_union" function is equal to expected

        <b>Steps</b>
        1. Check value returned by "bounding_box_intersection_over_union" function when "x_right" coordinate of
        intersection box more than "x_left"
        2. Check value returned by "bounding_box_intersection_over_union" function when "y_bottom" coordinate of
        intersection box more than "y_top"
        3. Check value returned by "bounding_box_intersection_over_union" function when boxes intersect in two points
        """
        # Checking value returned by "bounding_box_intersection_over_union" when "x_right" coordinate of
        # intersection box more than "x_left"
        assert bounding_box_intersection_over_union(box1=[2, 2, 5, 6], box2=[7, 4, 4, 8]) == 0.0
        # Checking value returned by "bounding_box_intersection_over_union" when "y_bottom" coordinate of
        # intersection box more than "y_top"
        assert bounding_box_intersection_over_union(box1=[2, 8, 6, 1], box2=[1, 7, 5, 2]) == 0.0
        # Checking value returned by "bounding_box_intersection_over_union" when boxes intersect in two points
        assert bounding_box_intersection_over_union(box1=[1, 3, 3, 7], box2=[2, 4, 5, 5]) == 0.1

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_get_iou_matrix(self):
        """
        <b>Description:</b>
        Check "get_iou_matrix" function

        <b>Input data:</b>
        Bounding boxes coordinates

        <b>Expected results:</b>
        Test passes if array returned by "get_iou_matrix" function is equal to expected
        """
        boxes_1 = [[2, 2, 5, 6], [2, 8, 6, 1], [1, 3, 3, 7]]
        boxes_2 = [[7, 4, 4, 8], [2, 4, 5, 5], [1, 1, 2, 2], [0, 0, 10, 10]]
        expected_matrix = [
            [0.0, 0.25, 0.0, 0.12],
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.1, 0.0, 0.08],
        ]
        assert np.array_equal(get_iou_matrix(boxes_1, boxes_2), expected_matrix)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_get_n_false_negatives(self):
        """
        <b>Description:</b>
        Check "get_n_false_negatives" function

        <b>Input data:</b>
        IoU-matrix np.array

        <b>Expected results:</b>
        Test passes if value returned by "get_n_false_negatives" function is equal to expected

        <b>Steps</b>
        1. Check value returned by "get_n_false_negatives" function when max element in a row is less than
        "iou_threshold" parameter
        2. Check value returned by "get_n_false_negatives" function when several elements in a column are more than
        "iou_threshold" parameter
        """
        iou_matrix = np.array([[0.0, 0.25, 0.0, 0.09], [0.0, 0.0, 0.0, 0.0], [0.0, 0.1, 0.0, 0.08]])
        # Checking value returned by "get_n_false_negatives" when max element in a row is less than "iou_threshold"
        assert get_n_false_negatives(iou_matrix, 0.11) == 2
        # Checking value returned by "get_n_false_negatives" when several elements in a column are more than
        # "iou_threshold"
        assert get_n_false_negatives(iou_matrix, 0.09) == 2


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestMetrics:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_metrics_initialization(self):
        """
        <b>Description:</b>
        Check "_Metrics" class object initialization

        <b>Input data:</b>
        "_Metrics" class object with specified "f_measure", "precision" and "recall" parameters

        <b>Expected results:</b>
        Test passes if "f_measure", "precision" and "recall" attributes of initialized "_Metrics" class object
        are equal to expected
        """
        f_measure = 0.4
        precision = 0.9
        recall = 0.3
        metrics = _Metrics(f_measure=f_measure, precision=precision, recall=recall)
        assert metrics.f_measure == f_measure
        assert metrics.precision == precision
        assert metrics.recall == recall


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestResultCounters:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_counters_initialization(self):
        """
        <b>Description:</b>
        Check "_ResultCounters" class object initialization

        <b>Input data:</b>
        "_ResultCounters" class object with specified "n_false_negatives", "n_true" and "n_predicted" parameters

        <b>Expected results:</b>
        Test passes if "n_false_negatives", "n_true" and "n_predicted" attributes of initialized "_ResultCounters"
        class object are equal to expected
        """
        n_false_negatives = 2
        n_true = 9
        n_predicted = 9
        result_counters = _ResultCounters(n_false_negatives=n_false_negatives, n_true=n_true, n_predicted=n_predicted)
        assert result_counters.n_false_negatives == n_false_negatives
        assert result_counters.n_true == n_true
        assert result_counters.n_predicted == n_predicted

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_counters_calculate_f_measure(self):
        """
        <b>Description:</b>
        Check "_ResultCounters" class "calculate_f_measure" method

        <b>Input data:</b>
        "_ResultCounters" class object with specified "n_false_negatives", "n_true" and "n_predicted" parameters

        <b>Expected results:</b>
        Test passes if value returned by "calculate_f_measure" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "calculate_f_measure" method when "n_true" attribute of "_ResultCounters" object is
        more than "n_false_negatives"
        2. Check value returned by "calculate_f_measure" method when "n_true" attribute of "_ResultCounters" object is
        less than "n_false_negatives"
        3. Check value returned by "calculate_f_measure" method when "n_true" attribute of "_ResultCounters" object is
        equal to "n_false_negatives"
        4. Check value returned by "calculate_f_measure" method when "n_predicted" attribute of "_ResultCounters" object
        is equal to 0
        5. Check value returned by "calculate_f_measure" method when "n_true" attribute of "_ResultCounters" object is
        equal to 0
        """

        def check_calculated_f_measure(
            f_measure_actual: _Metrics,
            expected_precision: float,
            expected_recall: float,
        ):
            expected_f_measure = (2 * expected_precision * expected_recall) / (
                expected_precision + expected_recall + np.finfo(float).eps
            )
            assert f_measure_actual.f_measure == expected_f_measure
            assert f_measure_actual.precision == expected_precision
            assert f_measure_actual.recall == expected_recall

        # Checking value returned by "calculate_f_measure" when "n_true" is more than "n_false_negatives"
        result_counters = _ResultCounters(n_false_negatives=4, n_true=8, n_predicted=16)
        precision = (8 - 4) / 16  # (n_true-n_false_negatives)/n_predicted
        recall = (8 - 4) / 8  # (n_true-n_false_negatives)/n_true
        f_measure = result_counters.calculate_f_measure()
        check_calculated_f_measure(f_measure, precision, recall)
        # Checking value returned by "calculate_f_measure" when "n_true" is less than "n_false_negatives"
        result_counters = _ResultCounters(n_false_negatives=16, n_true=8, n_predicted=16)
        f_measure = result_counters.calculate_f_measure()
        precision = (8 - 16) / 16  # (n_true-n_false_negatives)/n_predicted
        recall = (8 - 16) / 8  # (n_true-n_false_negatives)/n_true
        check_calculated_f_measure(f_measure, precision, recall)
        # Checking value returned by "calculate_f_measure" when "n_true" is equal to "n_false_negatives"
        result_counters = _ResultCounters(n_false_negatives=8, n_true=8, n_predicted=14)
        f_measure = result_counters.calculate_f_measure()
        check_calculated_f_measure(f_measure, 0.0, 0.0)
        # Checking value returned by "calculate_f_measure" when "n_predicted" is equal to 0
        result_counters = _ResultCounters(n_false_negatives=2, n_true=8, n_predicted=0)
        f_measure = result_counters.calculate_f_measure()
        check_calculated_f_measure(f_measure, 1.0, 0.0)
        # Checking value returned by "calculate_f_measure" method when "n_true" is equal to 0
        result_counters = _ResultCounters(n_false_negatives=2, n_true=0, n_predicted=8)
        f_measure = result_counters.calculate_f_measure()
        check_calculated_f_measure(f_measure, 0.0, 1.0)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestAggregatedResults:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_aggregate_results_initialization(self):
        """
        <b>Description:</b>
        Check "_AggregatedResults" class object initialization

        <b>Input data:</b>
        "_AggregatedResults" class object with specified "classes" parameter

        <b>Expected results:</b>
        Test passes if attributes of initialized "_AggregatedResults" class object are equal to expected

        <b>Steps</b>
        1. Check attributes of "_AggregatedResults" object initialized with "classes" attribute equal to list with
        several classes
        2. Check attributes of "_AggregatedResults" object initialized with "classes" attribute equal to list with one
        class
        3. Check attributes of "_AggregatedResults" object initialized with "classes" attribute equal to empty list
        """

        def check_aggregate_results_attributes(aggregate_results, expected_classes_curve):
            assert aggregate_results.f_measure_curve == expected_classes_curve
            assert aggregate_results.precision_curve == expected_classes_curve
            assert aggregate_results.recall_curve == expected_classes_curve
            assert aggregate_results.all_classes_f_measure_curve == []
            assert aggregate_results.best_f_measure == 0.0
            assert aggregate_results.best_threshold == 0.0

        # Checking attributes of "_AggregatedResults" object initialized with "classes" equal to list with
        # several classes
        check_aggregate_results_attributes(
            aggregate_results=_AggregatedResults(["class_1", "class_2", "class_3"]),
            expected_classes_curve={"class_1": [], "class_2": [], "class_3": []},
        )
        # Checking attributes of "_AggregatedResults" object initialized with "classes" equal to list with one class
        check_aggregate_results_attributes(
            aggregate_results=_AggregatedResults(["class_1"]),
            expected_classes_curve={"class_1": []},
        )
        # Checking attributes of "_AggregatedResults" object initialized with "classes" equal to empty list
        check_aggregate_results_attributes(aggregate_results=_AggregatedResults([]), expected_classes_curve={})


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestOverallResults:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_counters_initialization(self):
        """
        <b>Description:</b>
        Check "_OverallResults" class object initialization

        <b>Input data:</b>
        "_OverallResults" class object with specified "per_confidence", "per_nms", "best_f_measure_per_class" and
        "best_f_measure" parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "_OverallResults" class object are equal to expected
        """
        per_confidence = _AggregatedResults(["confidence_class_1", "confidence_class_2"])
        per_nms = _AggregatedResults(["nms_class_1", "nms_class_2"])
        best_f_measure_per_class = {"class_1": 0.6, "class_2": 0.8}
        best_f_measure = 0.8
        overall_results = _OverallResults(
            per_confidence=per_confidence,
            per_nms=per_nms,
            best_f_measure_per_class=best_f_measure_per_class,
            best_f_measure=best_f_measure,
        )
        assert overall_results.per_confidence == per_confidence
        assert overall_results.per_nms == per_nms
        assert overall_results.best_f_measure_per_class == best_f_measure_per_class
        assert overall_results.best_f_measure == best_f_measure


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestFMeasureCalculator:
    @staticmethod
    def ground_truth_boxes_per_image():
        return [  # images
            [  # image_1 boxes
                (0.5, 0.1, 0.8, 0.9, "class_1", 0.95),
                (0.1, 0.1, 0.3, 0.7, "class_2", 0.93),
                (0.05, 0.05, 0.3, 0.75, "class_2", 0.91),
            ],
            [  # image_2 boxes
                (0.15, 0.0, 0.4, 0.75, "class_1", 0.9),
                (0.1, 0.0, 0.45, 0.7, "class_1", 0.88),
                (0.45, 0.0, 0.95, 0.45, "class_3", 0.94),
                (0.45, 0.0, 1.0, 0.5, "class_3", 0.92),
            ],
        ]

    @staticmethod
    def prediction_boxes_per_image():
        return [  # images
            [  # image_1 boxes
                (0.45, 0.2, 0.75, 0.85, "class_1", 0.92),
                (0.5, 0.05, 0.8, 0.85, "class_1", 0.93),
                (0.1, 0.15, 0.35, 0.75, "class_2", 0.92),
                (0.05, 0.05, 0.35, 0.7, "class_2", 0.91),
            ],
            [  # image_2 boxes
                (0.15, 0.05, 0.45, 0.8, "class_1", 0.89),
                (0.15, 0.0, 0.5, 0.7, "class_1", 0.85),
                (0.5, 0.0, 0.9, 0.5, "class_3", 0.95),
                (0.45, 0.05, 0.95, 0.5, "class_3", 0.94),
            ],
        ]

    def f_measure_calculator(self):
        return _FMeasureCalculator(
            ground_truth_boxes_per_image=self.ground_truth_boxes_per_image(),
            prediction_boxes_per_image=self.prediction_boxes_per_image(),
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_initialization(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class object initialization

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "_FMeasureCalculator" class object are equal to expected
        """
        ground_truth_boxes_per_image = self.ground_truth_boxes_per_image()
        prediction_boxes_per_image = self.prediction_boxes_per_image()
        f_measure_calculator = _FMeasureCalculator(
            ground_truth_boxes_per_image=ground_truth_boxes_per_image,
            prediction_boxes_per_image=prediction_boxes_per_image,
        )
        assert f_measure_calculator.ground_truth_boxes_per_image == ground_truth_boxes_per_image
        assert f_measure_calculator.prediction_boxes_per_image == prediction_boxes_per_image
        assert f_measure_calculator.confidence_range == [0.025, 1.0, 0.025]
        assert f_measure_calculator.nms_range == [0.1, 1, 0.05]
        assert f_measure_calculator.default_confidence_threshold == 0.35

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_get_counters(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "get_counters" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if "_ResultCounters" object returned by "get_counters" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "get_counters" method for "ground_truth_boxes" and "predicted_boxes" attributes with
        length is more than 0
        2. Check value returned by "get_counters" method for "predicted_boxes" attribute with length is more than 0 and
        "ground_truth_boxes" attribute with length is equal to 0
        3. Check value returned by "get_counters" method for "ground_truth_boxes" attribute with length is more than 0
        and "predicted_boxes" attribute with length is equal to 0
        """

        def check_get_counters(
            calculator: _FMeasureCalculator,
            expected_n_false_negatives: int,
            expected_n_predicted: int,
            expected_n_true: int,
        ):
            get_counters = calculator.get_counters(0.75)
            assert isinstance(get_counters, _ResultCounters)
            assert get_counters.n_false_negatives == expected_n_false_negatives
            assert get_counters.n_predicted == expected_n_predicted
            assert get_counters.n_true == expected_n_true

        ground_boxes = self.ground_truth_boxes_per_image()
        predicted_boxes = self.prediction_boxes_per_image()
        # Checking value returned by "get_counters" for "ground_truth_boxes" and "predicted_boxes" with length is more
        # than 0
        f_measure_calculator = _FMeasureCalculator(
            ground_truth_boxes_per_image=ground_boxes,
            prediction_boxes_per_image=predicted_boxes,
        )
        check_get_counters(
            f_measure_calculator,
            expected_n_false_negatives=3,
            expected_n_predicted=8,
            expected_n_true=7,
        )
        # Checking value returned by "get_counters" for "predicted_boxes" with length is more than 0 and
        # "ground_truth_boxes" with length is equal to 0
        f_measure_calculator = _FMeasureCalculator(
            ground_truth_boxes_per_image=[[]],
            prediction_boxes_per_image=predicted_boxes,
        )
        check_get_counters(
            f_measure_calculator,
            expected_n_false_negatives=0,
            expected_n_predicted=4,
            expected_n_true=0,
        )
        # Checking value returned by "get_counters" for "ground_truth_boxes" with length is more than 0 and
        # "predicted_boxes" with length is equal to 0
        f_measure_calculator = _FMeasureCalculator(
            ground_truth_boxes_per_image=ground_boxes, prediction_boxes_per_image=[[]]
        )
        check_get_counters(
            f_measure_calculator,
            expected_n_false_negatives=3,
            expected_n_predicted=0,
            expected_n_true=3,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_filter_confidence(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "__filter_confidence" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object, "boxes_per_image" list and "confidence_threshold" parameter

        <b>Expected results:</b>
        Test passes if list returned by "__filter_confidence" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__filter_confidence" method for "confidence_threshold" equal to 0.0
        2. Check value returned by "__filter_confidence" method for "confidence_threshold" equal to filter some boxes
        3. Check value returned by "__filter_confidence" method for "confidence_threshold" equal to 1.0
        """
        f_measure_calculator = self.f_measure_calculator()
        boxes_per_image = f_measure_calculator.prediction_boxes_per_image
        # Checking value returned by "__filter_confidence" for "confidence_threshold" equal to 0.0
        assert (
            f_measure_calculator._FMeasureCalculator__filter_confidence(  # type: ignore[attr-defined]
                boxes_per_image, 0.0
            )
            == boxes_per_image
        )
        # Checking value returned by "__filter_confidence" for "confidence_threshold" equal to filter some boxes
        assert f_measure_calculator._FMeasureCalculator__filter_confidence(  # type: ignore[attr-defined]
            boxes_per_image, 0.92
        ) == [
            [(0.5, 0.05, 0.8, 0.85, "class_1", 0.93)],
            [
                (0.5, 0.0, 0.9, 0.5, "class_3", 0.95),
                (0.45, 0.05, 0.95, 0.5, "class_3", 0.94),
            ],
        ]
        assert f_measure_calculator._FMeasureCalculator__filter_confidence(  # type: ignore[attr-defined]
            boxes_per_image, 0.93
        ) == [
            [],
            [
                (0.5, 0.0, 0.9, 0.5, "class_3", 0.95),
                (0.45, 0.05, 0.95, 0.5, "class_3", 0.94),
            ],
        ]
        # Checking value returned by "__filter_confidence" for "confidence_threshold" equal to 1.0
        assert f_measure_calculator._FMeasureCalculator__filter_confidence(  # type: ignore[attr-defined]
            boxes_per_image, 1.0
        ) == [
            [],
            [],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_filter_class(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "__filter_class" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object, "boxes_per_image" list and classes to filter

        <b>Expected results:</b>
        Test passes if list returned by "__filter_class" method is equal to expected

        <b>Steps</b>
        1. Check list returned by "__filter_class" method to get class represented in one image
        2. Check list returned by "__filter_class" method to get class represented in several images
        3. Check list returned by "__filter_class" method to get class that is not represented in any of images
        """
        f_measure_calculator = self.f_measure_calculator()
        boxes_per_image = f_measure_calculator.prediction_boxes_per_image
        # Checking list returned by "__filter_class" to get class represented in one image
        assert f_measure_calculator._FMeasureCalculator__filter_class(  # type: ignore[attr-defined]
            boxes_per_image, "class_2"
        ) == [
            [
                (0.1, 0.15, 0.35, 0.75, "class_2", 0.92),
                (0.05, 0.05, 0.35, 0.7, "class_2", 0.91),
            ],
            [],
        ]
        # Checking list returned by "__filter_class" to get class represented in several images
        assert f_measure_calculator._FMeasureCalculator__filter_class(  # type: ignore[attr-defined]
            boxes_per_image, "class_1"
        ) == [
            [
                (0.45, 0.2, 0.75, 0.85, "class_1", 0.92),
                (0.5, 0.05, 0.8, 0.85, "class_1", 0.93),
            ],
            [
                (0.15, 0.05, 0.45, 0.8, "class_1", 0.89),
                (0.15, 0.0, 0.5, 0.7, "class_1", 0.85),
            ],
        ]
        # Checking list returned by "__filter_class" to get class that is not represented in any of images
        assert f_measure_calculator._FMeasureCalculator__filter_class(  # type: ignore[attr-defined]
            boxes_per_image, "class_6"
        ) == [
            [],
            [],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_filter_nms(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "__filter_nms" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object, "boxes_per_image" list, "critical_nms" list and "nms_threshold" values

        <b>Expected results:</b>
        Test passes if list returned by "__filter_nms" method is equal to expected

        <b>Steps</b>
        1. Check list returned by "__filter_nms" method for "nms_threshold" parameter that not filters any boxes
        2. Check list returned by "__filter_nms" method for "nms_threshold" that filters some boxes
        3. Check list returned by "__filter_nms" method for "nms_threshold" parameter that filters all boxes
        """
        f_measure_calculator = self.f_measure_calculator()
        boxes_per_image = f_measure_calculator.prediction_boxes_per_image
        critical_nms = [[0.5, 0.55, 0.65, 0.6], [0.6, 0.55, 0.5, 0.65]]
        # Checking list returned by "__filter_nms" for "nms_threshold" that not filters any boxes
        assert (
            f_measure_calculator._FMeasureCalculator__filter_nms(  # type: ignore[attr-defined]
                boxes_per_image, critical_nms, 1.0
            )
            == boxes_per_image
        )
        # Checking list returned by "__filter_nms" for "nms_threshold" that filters some boxes
        assert f_measure_calculator._FMeasureCalculator__filter_nms(  # type: ignore[attr-defined]
            boxes_per_image, critical_nms, 0.6
        ) == [
            [
                (0.45, 0.2, 0.75, 0.85, "class_1", 0.92),
                (0.5, 0.05, 0.8, 0.85, "class_1", 0.93),
            ],
            [
                (0.15, 0.0, 0.5, 0.7, "class_1", 0.85),
                (0.5, 0.0, 0.9, 0.5, "class_3", 0.95),
            ],
        ]
        # Checking list returned by "__filter_nms" for "nms_threshold" that filters all boxes
        assert f_measure_calculator._FMeasureCalculator__filter_nms(  # type: ignore[attr-defined]
            boxes_per_image, critical_nms, 0.1
        ) == [
            [],
            [],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_critical_nms(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "__get_critical_nms" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object, "boxes_per_image" list and "cross_class_nms" bool value

        <b>Expected results:</b>
        Test passes if list returned by "__get_critical_nms" method is equal to expected

        <b>Steps</b>
        1. Check list returned by "__get_critical_nms" method when "cross_class_nms" parameter is "False"
        2. Check list returned by "__get_critical_nms" method when "cross_class_nms" parameter is "True"
        """
        f_measure_calculator = self.f_measure_calculator()
        boxes_per_image = [  # images
            [  # image_1 boxes
                (0.1, 0.1, 0.4, 0.4, "class_1", 0.98),
                (0.2, 0.2, 0.5, 0.5, "class_1", 0.97),
                (0.6, 0.2, 0.9, 0.5, "class_2", 0.97),
                (0.6, 0.3, 1.0, 0.6, "class_3", 0.98),
            ],
            [  # image_2 boxes
                (0.1, 0.1, 0.4, 0.5, "class_2", 0.94),
                (0.2, 0.2, 0.5, 0.6, "class_2", 0.95),
                (0.6, 0.1, 0.9, 0.6, "class_3", 0.92),
                (0.7, 0.1, 1.0, 0.6, "class_4", 0.94),
            ],
        ]
        # Checking list returned by "__get_critical_nms" when "cross_class_nms" is "False"
        assert f_measure_calculator._FMeasureCalculator__get_critical_nms(  # type: ignore[attr-defined]
            boxes_per_image, False
        ) == [
            [0.0, 0.28571428571428575, 0.0, 0.0],
            [0.3333333333333333, 0.0, 0.0, 0.0],
        ]
        # Checking list returned by "__get_critical_nms" when "cross_class_nms" is "True"
        assert f_measure_calculator._FMeasureCalculator__get_critical_nms(  # type: ignore[attr-defined]
            boxes_per_image, True
        ) == [
            [0.0, 0.28571428571428575, 0.4, 0.0],
            [0.3333333333333333, 0.0, 0.5000000000000001, 0.0],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_get_f_measure_for_class(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "get_f_measure_for_class" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if tuple returned by "get_f_measure_for_class" method is equal to expected

        <b>Steps</b>
        1. Check tuple returned by "get_f_measure_for_class" method for class included in _FMeasureCalculator object
        2. Check tuple returned by "get_f_measure_for_class" method for class non-included in _FMeasureCalculator object
        3. Check tuple returned by "get_f_measure_for_class" method for _FMeasureCalculator object with
        "ground_truth_boxes_per_image" parameter equal to empty list
        """

        def check_get_f_measure(actual_f_measure_for_class, expected_values_dict: dict):
            # Checking _Metrics object
            metrics = actual_f_measure_for_class[0]
            assert isinstance(metrics, _Metrics)
            assert metrics.f_measure == expected_values_dict.get("f_measure")
            assert metrics.precision == expected_values_dict.get("precision")
            assert metrics.recall == expected_values_dict.get("recall")
            # Checking _ResultCounters object
            result_counters = actual_f_measure_for_class[1]
            assert isinstance(result_counters, _ResultCounters)
            assert result_counters.n_false_negatives == expected_values_dict.get("n_false_negatives")
            assert result_counters.n_predicted == expected_values_dict.get("n_predicted")
            assert result_counters.n_true == expected_values_dict.get("n_true")

        f_measure_calculator = self.f_measure_calculator()
        # Checking tuple returned by "get_f_measure_for_class" method for class included in _FMeasureCalculator object
        check_get_f_measure(
            actual_f_measure_for_class=f_measure_calculator.get_f_measure_for_class("class_1", 0.75, 0.9),
            expected_values_dict={
                "f_measure": 0.3999999999999999,
                "precision": 0.5,
                "recall": 0.3333333333333333,
                "n_false_negatives": 2,
                "n_predicted": 2,
                "n_true": 3,
            },
        )
        # Checking tuple returned by "get_f_measure_for_class" method for non-included class in
        # _FMeasureCalculator object
        check_get_f_measure(
            actual_f_measure_for_class=f_measure_calculator.get_f_measure_for_class("class_6", 0.75, 0.9),
            expected_values_dict={
                "f_measure": 0.0,
                "precision": 1.0,
                "recall": 0.0,
                "n_false_negatives": 0,
                "n_predicted": 0,
                "n_true": 0,
            },
        )
        # Checking tuple returned by "get_f_measure_for_class" method for _FMeasureCalculator object with empty list
        # of ground_truth_boxes_per_image
        f_measure_calculator.ground_truth_boxes_per_image = []
        check_get_f_measure(
            actual_f_measure_for_class=f_measure_calculator.get_f_measure_for_class("class_1", 0.75, 0.9),
            expected_values_dict={
                "f_measure": 0.0,
                "precision": 0.0,
                "recall": 0.0,
                "n_false_negatives": 0,
                "n_predicted": 0,
                "n_true": 0,
            },
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_evaluate_classes(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "evaluate_classes" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if dictionary returned by "evaluate_classes" method is equal to expected

        <b>Steps</b>
        1. Check dictionary returned by "evaluate_classes" method when one class is specified as "classes" parameter
        2. Check dictionary returned by "evaluate_classes" method when several classes specified as "classes" parameter
        3. Check dictionary returned by "evaluate_classes" method when "All classes" is specified in "classes" parameter
        """

        def compare_metrics(actual_metric: _Metrics, expected_metric: _Metrics):
            assert actual_metric.f_measure == expected_metric.f_measure
            assert actual_metric.recall == expected_metric.recall
            assert actual_metric.precision == expected_metric.precision

        f_measure_calculator = self.f_measure_calculator()
        # Checking dictionary returned by "evaluate_classes" when one class is specified as "classes" parameter
        evaluate_classes = f_measure_calculator.evaluate_classes(["class_1"], 0.87, 0.8)
        expected_class_1 = f_measure_calculator.get_f_measure_for_class("class_1", 0.87, 0.8)[0]
        compare_metrics(evaluate_classes.get("class_1"), expected_class_1)
        compare_metrics(evaluate_classes.get("All Classes"), expected_class_1)
        # Checking dictionary returned by "evaluate_classes" when several classes specified as "classes" parameter
        evaluate_classes = f_measure_calculator.evaluate_classes(["class_1", "class_2"], 0.87, 0.8)
        expected_class_1 = f_measure_calculator.get_f_measure_for_class("class_1", 0.87, 0.8)[0]
        expected_class_2 = f_measure_calculator.get_f_measure_for_class("class_2", 0.87, 0.8)[0]
        expected_all_classes = _ResultCounters(n_false_negatives=4, n_predicted=6, n_true=5).calculate_f_measure()
        compare_metrics(evaluate_classes.get("class_1"), expected_class_1)
        compare_metrics(evaluate_classes.get("class_2"), expected_class_2)
        compare_metrics(evaluate_classes.get("All Classes"), expected_all_classes)
        # Checking dictionary returned by "evaluate_classes" when "All classes" is specified in "classes" parameter
        evaluate_classes = f_measure_calculator.evaluate_classes(["class_1", "All Classes"], 0.87, 0.8)
        compare_metrics(evaluate_classes.get("class_1"), expected_class_1)
        compare_metrics(evaluate_classes.get("All Classes"), expected_class_1)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_get_results_per_nms(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "get_results_per_nms" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if "_AggregatedResults" object returned by "get_results_per_nms" method is equal to expected

        <b>Steps</b>
        1. Check "_AggregatedResults" object returned by "get_results_per_nms" method when "cross_class_nms" parameter
        is "False" and All Classes "f-measure" is less than "min_f_measure" parameter
        2. Check "_AggregatedResults" object returned by "get_results_per_nms" method when "cross_class_nms" parameter
        is "True" and All Classes "f-measure" is more than "min_f_measure" parameter
        """

        def check_critical_nms_per_image(
            calculator: _FMeasureCalculator,
            actual_results: _AggregatedResults,
            iou_threshold: float,
            cross_class_nms: bool,
            expected_best_f_measure: float,
            expected_best_threshold: float,
        ):
            exp_results_per_nms = _AggregatedResults(["class_1", "class_2"])
            exp_critical_nms_per_image = calculator._FMeasureCalculator__get_critical_nms(  # type: ignore[attr-defined]
                calculator.prediction_boxes_per_image, cross_class_nms
            )
            for nms_threshold in np.arange(*calculator.nms_range):
                predict_boxes_per_image_nms = calculator._FMeasureCalculator__filter_nms(  # type: ignore[attr-defined]
                    calculator.prediction_boxes_per_image,
                    exp_critical_nms_per_image,
                    nms_threshold,
                )
                boxes_pair_for_nms = _FMeasureCalculator(
                    calculator.ground_truth_boxes_per_image,
                    predict_boxes_per_image_nms,
                )
                result_point = boxes_pair_for_nms.evaluate_classes(
                    classes=["class_1", "class_2"],
                    iou_threshold=iou_threshold,
                    confidence_threshold=calculator.default_confidence_threshold,
                )
                all_classes_f_measure = result_point["All Classes"].f_measure
                exp_results_per_nms.all_classes_f_measure_curve.append(all_classes_f_measure)

                for class_name in ["class_1", "class_2"]:
                    exp_results_per_nms.f_measure_curve[class_name].append(result_point[class_name].f_measure)
                    exp_results_per_nms.precision_curve[class_name].append(result_point[class_name].precision)
                    exp_results_per_nms.recall_curve[class_name].append(result_point[class_name].recall)
            assert actual_results.all_classes_f_measure_curve == exp_results_per_nms.all_classes_f_measure_curve
            assert actual_results.f_measure_curve == exp_results_per_nms.f_measure_curve
            assert actual_results.precision_curve == exp_results_per_nms.precision_curve
            assert actual_results.recall_curve == exp_results_per_nms.recall_curve
            assert actual_results.best_f_measure == expected_best_f_measure
            assert actual_results.best_threshold == expected_best_threshold

        f_measure_calculator = self.f_measure_calculator()
        # Checking "_AggregatedResults" object returned by "get_results_per_nms" when "cross_class_nms" is "False" and
        # All Classes "f-measure" is more than "min_f_measure"
        actual_results_per_nms = f_measure_calculator.get_results_per_nms(
            classes=["class_1", "class_2"], iou_threshold=0.6, min_f_measure=0.5
        )
        check_critical_nms_per_image(
            f_measure_calculator,
            actual_results_per_nms,
            iou_threshold=0.6,
            cross_class_nms=False,
            expected_best_f_measure=0.7499999999999998,
            expected_best_threshold=0.5500000000000002,
        )
        # Checking "_AggregatedResults" object returned by "get_results_per_nms" when "cross_class_nms" is "True" and
        # All Classes "f-measure" is less than "min_f_measure"
        actual_results_per_nms = f_measure_calculator.get_results_per_nms(
            classes=["class_1", "class_2"],
            iou_threshold=0.7,
            min_f_measure=0.8,
            cross_class_nms=True,
        )
        check_critical_nms_per_image(
            f_measure_calculator,
            actual_results_per_nms,
            iou_threshold=0.7,
            cross_class_nms=True,
            expected_best_f_measure=0.8,
            expected_best_threshold=0.5,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_get_results_per_confidence(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "get_results_per_confidence" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object with specified "ground_truth_boxes_per_image" and
        "prediction_boxes_per_image" parameters

        <b>Expected results:</b>
        Test passes if "_AggregatedResults" object returned by "get_results_per_confidence" method is equal to expected

        <b>Steps</b>
        1. Check "_AggregatedResults" object returned by "get_results_per_confidence" method when All Classes f-measure
        is more than best f-measure in results_per_confidence
        2. Check "_AggregatedResults" object returned by "get_results_per_confidence" method when All Classes f-measure
        is less than best f-measure in results_per_confidence
        """
        f_measure_calculator = self.f_measure_calculator()
        # Check "_AggregatedResults" object returned by "get_results_per_confidence" when All Classes f-measure is more
        # than best f-measure in results_per_confidence
        expected_results_per_confidence = _AggregatedResults(["class_1", "class_2"])
        for confidence_threshold in np.arange(*[0.6, 0.9]):
            result_point = f_measure_calculator.evaluate_classes(
                classes=["class_1", "class_2"],
                iou_threshold=0.7,
                confidence_threshold=confidence_threshold,
            )
            all_classes_f_measure = result_point["All Classes"].f_measure
            expected_results_per_confidence.all_classes_f_measure_curve.append(all_classes_f_measure)

            for class_name in ["class_1", "class_2"]:
                expected_results_per_confidence.f_measure_curve[class_name].append(result_point[class_name].f_measure)
                expected_results_per_confidence.precision_curve[class_name].append(result_point[class_name].precision)
                expected_results_per_confidence.recall_curve[class_name].append(result_point[class_name].recall)

        actual_results_per_confidence = f_measure_calculator.get_results_per_confidence(
            classes=["class_1", "class_2"],
            confidence_range=[0.6, 0.9],
            iou_threshold=0.7,
        )
        assert actual_results_per_confidence.all_classes_f_measure_curve == (
            expected_results_per_confidence.all_classes_f_measure_curve
        )
        assert actual_results_per_confidence.f_measure_curve == expected_results_per_confidence.f_measure_curve
        assert actual_results_per_confidence.recall_curve == expected_results_per_confidence.recall_curve
        assert actual_results_per_confidence.best_f_measure == 0.5454545454545453
        assert actual_results_per_confidence.best_threshold == 0.6
        # Check "_AggregatedResults" object returned by "get_results_per_confidence" when All Classes f-measure is less
        # than best f-measure in results_per_confidence
        actual_results_per_confidence = f_measure_calculator.get_results_per_confidence(
            classes=["class_1", "class_2"],
            confidence_range=[0.6, 0.9],
            iou_threshold=1.0,
        )
        assert actual_results_per_confidence.all_classes_f_measure_curve == [0.0]
        assert actual_results_per_confidence.f_measure_curve == {
            "class_1": [0.0],
            "class_2": [0.0],
        }
        assert actual_results_per_confidence.recall_curve == {
            "class_1": [0.0],
            "class_2": [0.0],
        }
        assert actual_results_per_confidence.best_f_measure == 0.0
        assert actual_results_per_confidence.best_threshold == 0.1

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_calculator_evaluate_detections(self):
        """
        <b>Description:</b>
        Check "_FMeasureCalculator" class "evaluate_detections" method

        <b>Input data:</b>
        "_FMeasureCalculator" class object, "boxes_per_image" list and "cross_class_nms" bool value

        <b>Expected results:</b>
        Test passes if "_OverallResults" object returned by "evaluate_detections" method is equal to expected

        <b>Steps</b>
        1. Check "_OverallResults" object returned by "evaluate_detections" method with default optional parameters
        2. Check "_OverallResults" object returned by "evaluate_detections" method with specified optional parameters
        """

        def check_evaluate_detections(
            calculator: _FMeasureCalculator,
            evaluate_detection: _OverallResults,
            iou_threshold: float = 0.5,
            result_based_nms_threshold: bool = False,
            cross_class_nms: bool = True,
        ):
            best_f_measure_per_class = {}
            results_per_confidence = calculator.get_results_per_confidence(
                classes=["class_1", "class_3"],
                confidence_range=calculator.confidence_range,
                iou_threshold=iou_threshold,
            )
            best_f_measure = results_per_confidence.best_f_measure
            for class_name in ["class_1", "class_3"]:
                best_f_measure_per_class[class_name] = max(results_per_confidence.f_measure_curve[class_name])
            results_per_nms = None
            if result_based_nms_threshold:
                results_per_nms = calculator.get_results_per_nms(
                    classes=["class_1", "class_3"],
                    iou_threshold=iou_threshold,
                    min_f_measure=results_per_confidence.best_f_measure,
                    cross_class_nms=cross_class_nms,
                )

                for class_name in ["class_1", "class_3"]:
                    best_f_measure_per_class[class_name] = max(results_per_nms.f_measure_curve[class_name])
            expected_evaluate_detection = _OverallResults(
                results_per_confidence,
                results_per_nms,
                best_f_measure_per_class,
                best_f_measure,
            )
            assert isinstance(evaluate_detection, _OverallResults)
            assert evaluate_detection.best_f_measure == expected_evaluate_detection.best_f_measure
            assert evaluate_detection.best_f_measure_per_class == expected_evaluate_detection.best_f_measure_per_class
            assert evaluate_detection.per_confidence.all_classes_f_measure_curve == (
                expected_evaluate_detection.per_confidence.all_classes_f_measure_curve
            )
            assert evaluate_detection.per_confidence.best_f_measure == (
                expected_evaluate_detection.per_confidence.best_f_measure
            )
            assert evaluate_detection.per_confidence.best_threshold == (
                expected_evaluate_detection.per_confidence.best_threshold
            )
            assert evaluate_detection.per_confidence.f_measure_curve == (
                expected_evaluate_detection.per_confidence.f_measure_curve
            )
            assert evaluate_detection.per_confidence.precision_curve == (
                expected_evaluate_detection.per_confidence.precision_curve
            )
            assert evaluate_detection.per_confidence.recall_curve == (
                expected_evaluate_detection.per_confidence.recall_curve
            )

        f_measure_calculator = self.f_measure_calculator()
        # Checking "_OverallResults" object returned by "evaluate_detections" with default optional parameters
        actual_evaluate_detection = f_measure_calculator.evaluate_detections(["class_1", "class_3"])
        check_evaluate_detections(
            calculator=f_measure_calculator,
            evaluate_detection=actual_evaluate_detection,
        )
        # Checking "_OverallResults" object returned by "evaluate_detections" with specified optional parameters
        actual_evaluate_detection = f_measure_calculator.evaluate_detections(["class_1", "class_3"], 0.79, True, True)
        check_evaluate_detections(
            calculator=f_measure_calculator,
            evaluate_detection=actual_evaluate_detection,
            iou_threshold=0.79,
            result_based_nms_threshold=True,
            cross_class_nms=True,
        )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestFMeasure:
    color = Color(0, 255, 0)
    creation_date = datetime.datetime(year=2021, month=12, day=23)

    @staticmethod
    def generate_random_image() -> Image:
        image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
        return image

    def model_labels(self):
        class_1_label = LabelEntity(
            name="class_1",
            domain=Domain.DETECTION,
            color=self.color,
            creation_date=self.creation_date,
        )
        class_2_label = LabelEntity(
            name="class_2",
            domain=Domain.DETECTION,
            color=self.color,
            creation_date=self.creation_date,
        )
        class_3_label = LabelEntity(
            name="class_3",
            domain=Domain.DETECTION,
            color=self.color,
            creation_date=self.creation_date,
        )
        return [class_1_label, class_2_label, class_3_label]

    def model(self):
        configurable_params = ConfigurableParameters(header="Test model configurable params")

        model_label_group = LabelGroup(name="model_labels", labels=self.model_labels(), id=ID("model_label"))

        model_configuration = ModelConfiguration(
            configurable_params, LabelSchemaEntity(label_groups=[model_label_group])
        )

        model = ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)
        return model

    def roi(self):
        return Annotation(
            shape=Rectangle(x1=0.0, y1=0.0, x2=1.0, y2=1.0),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="image_roi",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("image_roi"),
                    )
                )
            ],
        )

    def image_1_ground_boxes(self):
        box_1_annotation = Annotation(
            shape=Rectangle(x1=0.5, y1=0.1, x2=0.8, y2=0.9),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_1"),
                    ),
                    probability=0.95,
                )
            ],
        )

        box_2_annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0.1, x2=0.3, y2=0.7),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_2",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_2_image_1"),
                    ),
                    probability=0.93,
                )
            ],
        )

        box_3_annotation = Annotation(
            shape=Rectangle(x1=0.05, y1=0.05, x2=0.3, y2=0.75),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_2",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_2_image_1"),
                    ),
                    probability=0.91,
                )
            ],
        )

        annotation_scene = AnnotationSceneEntity(
            annotations=[box_1_annotation, box_2_annotation, box_3_annotation],
            kind=AnnotationSceneKind.ANNOTATION,
        )

        image_1 = DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=annotation_scene,
            roi=self.roi(),
        )
        return image_1

    def image_2_ground_boxes(self):
        box_1_annotation = Annotation(
            shape=Rectangle(x1=0.15, y1=0.0, x2=0.4, y2=0.75),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_2"),
                    ),
                    probability=0.9,
                )
            ],
        )

        box_2_annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0.0, x2=0.45, y2=0.7),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_2"),
                    ),
                    probability=0.88,
                )
            ],
        )

        box_3_annotation = Annotation(
            shape=Ellipse(x1=0.45, y1=0.0, x2=0.95, y2=0.45),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_3",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_3_image_2"),
                    ),
                    probability=0.94,
                )
            ],
        )

        box_4_annotation = Annotation(
            shape=Ellipse(x1=0.45, y1=0.0, x2=1.0, y2=0.5),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_3",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_3_image_2"),
                    ),
                    probability=0.92,
                )
            ],
        )

        annotation_scene = AnnotationSceneEntity(
            annotations=[
                box_1_annotation,
                box_2_annotation,
                box_3_annotation,
                box_4_annotation,
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )

        image_2 = DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=annotation_scene,
            roi=self.roi(),
        )
        return image_2

    def image_1_prediction_boxes(self):
        box_1_annotation = Annotation(
            shape=Rectangle(x1=0.45, y1=0.2, x2=0.75, y2=0.85),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_1"),
                    ),
                    probability=0.92,
                )
            ],
        )

        box_2_annotation = Annotation(
            shape=Rectangle(x1=0.5, y1=0.05, x2=0.8, y2=0.85),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_1"),
                    ),
                    probability=0.93,
                )
            ],
        )

        box_3_annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0.15, x2=0.35, y2=0.75),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_2",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_2_image_1"),
                    ),
                    probability=0.92,
                )
            ],
        )

        box_4_annotation = Annotation(
            shape=Rectangle(x1=0.05, y1=0.05, x2=0.35, y2=0.7),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_2",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_2_image_1"),
                    ),
                    probability=0.91,
                )
            ],
        )

        annotation_scene = AnnotationSceneEntity(
            annotations=[
                box_1_annotation,
                box_2_annotation,
                box_3_annotation,
                box_4_annotation,
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )

        image_1 = DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=annotation_scene,
            roi=self.roi(),
        )
        return image_1

    def image_2_prediction_boxes(self):
        box_1_annotation = Annotation(
            shape=Rectangle(x1=0.15, y1=0.05, x2=0.45, y2=0.8),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_2"),
                    ),
                    probability=0.89,
                )
            ],
        )

        box_2_annotation = Annotation(
            shape=Rectangle(x1=0.15, y1=0.0, x2=0.5, y2=0.7),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_1",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_1_image_2"),
                    ),
                    probability=0.85,
                )
            ],
        )

        box_3_annotation = Annotation(
            shape=Ellipse(x1=0.5, y1=0.0, x2=0.9, y2=0.5),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_3",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_3_image_2"),
                    ),
                    probability=0.95,
                )
            ],
        )

        box_4_annotation = Annotation(
            shape=Ellipse(x1=0.45, y1=0.05, x2=0.95, y2=0.5),
            labels=[
                ScoredLabel(
                    LabelEntity(
                        name="class_3",
                        domain=Domain.DETECTION,
                        color=self.color,
                        creation_date=self.creation_date,
                        id=ID("class_3_image_2"),
                    ),
                    probability=0.94,
                )
            ],
        )

        annotation_scene = AnnotationSceneEntity(
            annotations=[
                box_1_annotation,
                box_2_annotation,
                box_3_annotation,
                box_4_annotation,
            ],
            kind=AnnotationSceneKind.ANNOTATION,
        )

        image_2 = DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=annotation_scene,
            roi=self.roi(),
        )
        return image_2

    def ground_truth_dataset(self):
        return DatasetEntity([self.image_1_ground_boxes(), self.image_2_ground_boxes()])

    def prediction_dataset(self):
        return DatasetEntity([self.image_1_prediction_boxes(), self.image_2_prediction_boxes()])

    def incorrect_prediction_dataset(self):
        return DatasetEntity([self.image_2_prediction_boxes(), self.image_2_prediction_boxes()])

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_initialization(self):
        """
        <b>Description:</b>
        Check "FMeasure" class object initialization

        <b>Input data:</b>
        "FMeasure" class object with specified "ground_truth_boxes_per_image" and "prediction_boxes_per_image"
        parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "FMeasure" object are equal to expected

        <b>Steps</b>
        1. Check attributes of "FMeasure" class object initialized with default optional parameters
        2. Check attributes of "FMeasure" class object initialized with specified optional parameters
        3. Check ValueError exception is raised when empty list "ground_truth_dataset" is specified in "result_set"
        parameter
        4. Check ValueError exception is raised when empty list "prediction_dataset" is specified in "result_set"
        parameter
        """

        def check_f_measure_common_attributes(f_measure_actual):
            assert f_measure_actual.box_class_index == 4
            assert f_measure_actual.box_score_index == 5
            assert isinstance(f_measure_actual.f_measure, ScoreMetric)
            assert f_measure_actual.f_measure.name == "f-measure"
            assert f_measure_actual.f_measure.value == pytest.approx(0.2857142857142856)

        ground_dataset = self.ground_truth_dataset()
        prediction_dataset = self.prediction_dataset()
        result_set = ResultSetEntity(
            model=self.model(),
            ground_truth_dataset=ground_dataset,
            prediction_dataset=prediction_dataset,
        )
        labels = self.model_labels()
        # Checking attributes of "FMeasure" class object initialized with default optional parameters
        f_measure = FMeasure(result_set)
        check_f_measure_common_attributes(f_measure_actual=f_measure)
        assert f_measure.f_measure_per_label == {
            labels[0]: ScoreMetric(name="class_1", value=0.6666666666666665),
            labels[2]: ScoreMetric(name="class_3", value=0.0),
            labels[1]: ScoreMetric(name="class_2", value=0.0),
        }
        assert not f_measure.best_confidence_threshold
        assert not f_measure.best_nms_threshold
        assert not f_measure.f_measure_per_confidence
        assert not f_measure.f_measure_per_nms
        # Checking attributes of "FMeasure" class object initialized with specified optional parameters
        f_measure = FMeasure(
            resultset=result_set,
            vary_confidence_threshold=True,
            vary_nms_threshold=True,
            cross_class_nms=True,
        )
        check_f_measure_common_attributes(f_measure_actual=f_measure)
        assert f_measure.f_measure_per_label == {
            labels[0]: ScoreMetric(name="class_1", value=0.7999999999999999),
            labels[1]: ScoreMetric(name="class_2", value=0.6666666666666665),
            labels[2]: ScoreMetric(name="class_3", value=0.6666666666666665),
        }
        label_schema_labels = result_set.model.configuration.get_label_schema().get_labels(include_empty=False)
        classes = [label.name for label in label_schema_labels]
        boxes_pair = _FMeasureCalculator(
            f_measure._FMeasure__get_boxes_from_dataset_as_list(ground_dataset, labels),  # type: ignore[attr-defined]
            f_measure._FMeasure__get_boxes_from_dataset_as_list(  # type: ignore[attr-defined]
                prediction_dataset, labels
            ),
        )
        result = boxes_pair.evaluate_detections(
            result_based_nms_threshold=True,
            classes=classes,
            cross_class_nms=True,
        )
        expected_f_measure_per_confidence = CurveMetric(
            name="f-measure per confidence",
            xs=list(np.arange(*boxes_pair.confidence_range)),
            ys=result.per_confidence.all_classes_f_measure_curve,
        )
        assert f_measure.f_measure_per_confidence.name == expected_f_measure_per_confidence.name
        assert f_measure.f_measure_per_confidence.xs == expected_f_measure_per_confidence.xs
        assert f_measure.f_measure_per_confidence.ys == expected_f_measure_per_confidence.ys
        expected_f_measure_per_nms = CurveMetric(
            name="f-measure per nms",
            xs=list(np.arange(*boxes_pair.nms_range)),
            ys=result.per_nms.all_classes_f_measure_curve,
        )
        assert f_measure.f_measure_per_nms.name == expected_f_measure_per_nms.name
        assert f_measure.f_measure_per_nms.xs == expected_f_measure_per_nms.xs
        assert f_measure.f_measure_per_nms.ys == expected_f_measure_per_nms.ys
        # Checking ValueError exception is raised when empty list "ground_truth_dataset" is specified in "result_set"
        empty_dataset = DatasetEntity([])
        result_set = ResultSetEntity(
            model=self.model(),
            ground_truth_dataset=empty_dataset,
            prediction_dataset=prediction_dataset,
        )
        with pytest.raises(ValueError):
            FMeasure(result_set)
        # Checking ValueError exception is raised when empty list "prediction_dataset" is specified in "result_set"
        result_set = ResultSetEntity(
            model=self.model(),
            ground_truth_dataset=ground_dataset,
            prediction_dataset=empty_dataset,
        )
        with pytest.raises(ValueError):
            FMeasure(result_set)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_f_measure_get_performance(self):
        """
        <b>Description:</b>
        Check "FMeasure" class "get_performance" method

        <b>Input data:</b>
        "FMeasure" class object with specified "ground_truth_boxes_per_image" and "prediction_boxes_per_image"
        parameters

        <b>Expected results:</b>
        Test passes if "Performance" object returned by "get_performance" method is equal to expected

        <b>Steps</b>
        1. Check "Performance" object returned by "get_performance" method for "FMeasure" object initialized with
        default optional parameters
        2. Check "Performance" object returned by "get_performance" method for "FMeasure" object initialized with
        specified optional parameters
        """

        def check_performance(performance, expected_score, expected_metric_groups):
            assert isinstance(performance, Performance)
            assert isinstance(performance.score, ScoreMetric)
            assert performance.score.name == "f-measure"
            assert performance.score.value == pytest.approx(expected_score)
            # Checking dashboard metrics
            for expected_metric_group in expected_metric_groups:
                metric_group_index = expected_metric_groups.index(expected_metric_group)
                actual_metric_group = performance.dashboard_metrics[metric_group_index]
                if isinstance(expected_metric_group, BarMetricsGroup):
                    assert actual_metric_group.metrics == expected_metric_group.metrics
                    assert isinstance(actual_metric_group.visualization_info, BarChartInfo)
                    assert actual_metric_group.visualization_info.name == "F-measure per label"
                    assert actual_metric_group.visualization_info.palette == ColorPalette.LABEL
                    assert actual_metric_group.visualization_info.type == VisualizationType.RADIAL_BAR
                if isinstance(expected_metric_group, LineMetricsGroup):
                    assert actual_metric_group.metrics == expected_metric_group.metrics
                    assert actual_metric_group.visualization_info.name == expected_metric_group.visualization_info.name
                    assert actual_metric_group.visualization_info.palette == ColorPalette.DEFAULT
                    assert (
                        actual_metric_group.visualization_info.x_axis_label
                        == expected_metric_group.visualization_info.x_axis_label
                    )
                    assert (
                        actual_metric_group.visualization_info.y_axis_label
                        == expected_metric_group.visualization_info.y_axis_label
                    )
                if isinstance(expected_metric_group, TextMetricsGroup):
                    assert actual_metric_group.metrics == expected_metric_group.metrics
                    assert actual_metric_group.visualization_info.name == expected_metric_group.visualization_info.name
                    assert actual_metric_group.visualization_info.palette == ColorPalette.DEFAULT
                    assert actual_metric_group.visualization_info.type == VisualizationType.TEXT

        def generate_expected_default_dashboard_metric_groups(
            actual_f_measure: FMeasure,
        ):
            return [
                BarMetricsGroup(
                    metrics=list(actual_f_measure.f_measure_per_label.values()),
                    visualization_info=BarChartInfo(
                        name="F-measure per label",
                        palette=ColorPalette.LABEL,
                        visualization_type=VisualizationType.RADIAL_BAR,
                    ),
                )
            ]

        def generate_expected_optional_dashboard_metric_groups(
            actual_f_measure: FMeasure,
        ):
            return [
                generate_expected_default_dashboard_metric_groups(actual_f_measure)[0],
                LineMetricsGroup(
                    metrics=[cast(CurveMetric, actual_f_measure.f_measure_per_confidence)],
                    visualization_info=LineChartInfo(
                        name="F-measure per confidence",
                        x_axis_label="Confidence threshold",
                        y_axis_label="F-measure",
                    ),
                ),
                TextMetricsGroup(
                    metrics=[cast(ScoreMetric, actual_f_measure.best_confidence_threshold)],
                    visualization_info=TextChartInfo(name="Optimal confidence threshold"),
                ),
                LineMetricsGroup(
                    metrics=[cast(CurveMetric, actual_f_measure.f_measure_per_nms)],
                    visualization_info=LineChartInfo(
                        name="F-measure per nms",
                        x_axis_label="NMS threshold",
                        y_axis_label="F-measure",
                    ),
                ),
                TextMetricsGroup(
                    metrics=[cast(ScoreMetric, actual_f_measure.best_nms_threshold)],
                    visualization_info=TextChartInfo(name="Optimal nms threshold"),
                ),
            ]

        ground_dataset = self.ground_truth_dataset()
        prediction_dataset = self.prediction_dataset()
        result_set = ResultSetEntity(
            model=self.model(),
            ground_truth_dataset=ground_dataset,
            prediction_dataset=prediction_dataset,
        )
        # Checking "Performance" object returned by "get_performance" for "FMeasure" object initialized with default
        # optional parameters
        f_measure = FMeasure(result_set)
        expected_dashboard_metric_groups = generate_expected_default_dashboard_metric_groups(f_measure)
        actual_performance = f_measure.get_performance()
        check_performance(
            performance=actual_performance,
            expected_score=0.2857142857142856,
            expected_metric_groups=expected_dashboard_metric_groups,
        )
        # Check for incorrect prediction dataset
        incorrect_prediction_dataset = self.incorrect_prediction_dataset()
        incorrect_result_set = ResultSetEntity(
            model=self.model(),
            ground_truth_dataset=ground_dataset,
            prediction_dataset=incorrect_prediction_dataset,
        )
        f_measure = FMeasure(incorrect_result_set)
        expected_dashboard_metric_groups = generate_expected_default_dashboard_metric_groups(f_measure)
        actual_performance = f_measure.get_performance()
        check_performance(
            performance=actual_performance,
            expected_score=0.15384615384615372,
            expected_metric_groups=expected_dashboard_metric_groups,
        )
        # Checking attributes of "FMeasure" class object initialized with specified values of optional parameters
        f_measure = FMeasure(
            resultset=result_set,
            vary_confidence_threshold=True,
            vary_nms_threshold=True,
            cross_class_nms=True,
        )
        expected_dashboard_metric_groups = generate_expected_optional_dashboard_metric_groups(f_measure)
        actual_performance = f_measure.get_performance()
        check_performance(
            performance=actual_performance,
            expected_score=0.2857142857142856,
            expected_metric_groups=expected_dashboard_metric_groups,
        )
        # Check for incorrect prediction dataset
        f_measure = FMeasure(
            resultset=incorrect_result_set,
            vary_confidence_threshold=True,
            vary_nms_threshold=True,
            cross_class_nms=True,
        )
        expected_dashboard_metric_groups = generate_expected_optional_dashboard_metric_groups(f_measure)
        actual_performance = f_measure.get_performance()
        check_performance(
            performance=actual_performance,
            expected_score=0.15384615384615372,
            expected_metric_groups=expected_dashboard_metric_groups,
        )



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import tempfile

import openvino.runtime as ov
import pytest

from otx.core.ov.omz_wrapper import get_omz_model
from otx.core.ov.ops.infrastructures import ParameterV0
from otx.core.ov.ops.modules.op_module import convert_op_to_torch_module
from otx.core.ov.ops.utils import convert_op_to_torch
from otx.core.ov.utils import (
    get_op_name,
    load_ov_model,
    normalize_name,
    unnormalize_name,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_load_ov_model():
    with tempfile.TemporaryDirectory() as tempdir:
        model = get_omz_model("mobilenet-v2-pytorch", download_dir=tempdir, output_dir=tempdir)
        model = load_ov_model(**model)

        model = load_ov_model("omz://mobilenet-v2-pytorch")
        assert not model.inputs[0].get_partial_shape().is_dynamic

        model = load_ov_model("omz://mobilenet-v2-pytorch", convert_dynamic=True)
        assert model.inputs[0].get_partial_shape().is_dynamic


@e2e_pytest_unit
def test_normalize_name():
    assert "dummy" == normalize_name("dummy")
    assert "dummy#dummy" == normalize_name("dummy.dummy")
    assert "dummy###dummy" == normalize_name("dummy...dummy")
    assert "dummy#dummy#dummy" == normalize_name("dummy.dummy.dummy")


@e2e_pytest_unit
def test_unnormalize_name():
    assert "dummy" == unnormalize_name("dummy")
    assert "dummy.dummy" == unnormalize_name("dummy#dummy")
    assert "dummy...dummy" == unnormalize_name("dummy###dummy")
    assert "dummy.dummy.dummy" == unnormalize_name("dummy#dummy#dummy")


@e2e_pytest_unit
def test_get_op_name():
    assert "dummy" == get_op_name(ov.opset10.parameter([3, 1, 2], ov.Type.f32, name="dummy"))
    assert "dummy#dummy" == get_op_name(ov.opset10.parameter([3, 1, 2], ov.Type.f32, name="dummy.dummy"))


@e2e_pytest_unit
def test_convert_op_to_torch():
    dummy = ov.opset10.parameter([3, 1, 2], ov.Type.f32)
    assert isinstance(convert_op_to_torch(dummy), ParameterV0)

    dummy = ov.opset10.depth_to_space(dummy, mode="blocks_first")
    with pytest.raises(KeyError):
        convert_op_to_torch(dummy)


@e2e_pytest_unit
def test_convert_op_to_torch_module():
    data = ov.opset10.parameter([3, 1, 2], ov.Type.f32)
    mul_constant = ov.opset10.constant([1.5], ov.Type.f32, name="weight")
    mul = ov.opset10.multiply(data, mul_constant)

    module = convert_op_to_torch_module(mul)
    should_none, node = list(module._dependent_ops.values())
    assert should_none is None
    assert node is not None and node.name == "weight"



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=unused-argument
"""Code in this file is adapted from.

https://github.com/ildoonet/pytorch-randaugment/blob/master/RandAugment/augmentations.py
https://github.com/google-research/fixmatch/blob/master/third_party/auto_augment/augmentations.py
https://github.com/google-research/fixmatch/blob/master/libml/ctaugment.py
"""

import random

import numpy as np
import PIL
from mmcls.datasets.builder import PIPELINES

PARAMETER_MAX = 10


def auto_contrast(img, **kwargs):
    """Applies auto contrast to an image."""
    return PIL.ImageOps.autocontrast(img), None


def brightness(img, value, max_value, bias=0):
    """Applies brightness adjustment to an image."""
    value = _float_parameter(value, max_value) + bias
    return PIL.ImageEnhance.Brightness(img).enhance(value), value


def color(img, value, max_value, bias=0):
    """Applies color adjustment to an image."""
    value = _float_parameter(value, max_value) + bias
    return PIL.ImageEnhance.Color(img).enhance(value), value


def contrast(img, value, max_value, bias=0):
    """Applies contrast adjustment to an image."""
    value = _float_parameter(value, max_value) + bias
    return PIL.ImageEnhance.Contrast(img).enhance(value), value


def cutout(img, value, max_value, bias=0):
    """Applies cutout augmentation to an image."""
    if value == 0:
        return img
    value = _float_parameter(value, max_value) + bias
    value = int(value * min(img.size))
    return cutout_abs(img, value), value


def cutout_abs(img, value, **kwargs):
    """Applies cutout with absolute pixel size to an image."""
    w, h = img.size
    x0 = np.random.uniform(0, w)
    y0 = np.random.uniform(0, h)
    x0 = int(max(0, x0 - value / 2.0))
    y0 = int(max(0, y0 - value / 2.0))
    x1 = int(min(w, x0 + value))
    y1 = int(min(h, y0 + value))
    xy = (x0, y0, x1, y1)
    # gray
    rec_color = (127, 127, 127)
    img = img.copy()
    PIL.ImageDraw.Draw(img).rectangle(xy, rec_color)
    return img, xy, rec_color


def equalize(img, **kwargs):
    """Applies equalization to an image."""
    return PIL.ImageOps.equalize(img), None


def identity(img, **kwargs):
    """Returns the original image without any transformation."""
    return img, None


def posterize(img, value, max_value, bias=0):
    """Applies posterization to an image."""
    value = _int_parameter(value, max_value) + bias
    return PIL.ImageOps.posterize(img, value), value


def rotate(img, value, max_value, bias=0):
    """Applies rotation to an image."""
    value = _int_parameter(value, max_value) + bias
    if random.random() < 0.5:
        value = -value
    return img.rotate(value), value


def sharpness(img, value, max_value, bias=0):
    """Applies Sharpness to an image."""
    value = _float_parameter(value, max_value) + bias
    return PIL.ImageEnhance.Sharpness(img).enhance(value), value


def shear_x(img, value, max_value, bias=0):
    """Applies ShearX to an image."""
    value = _float_parameter(value, max_value) + bias
    if random.random() < 0.5:
        value = -value
    return img.transform(img.size, PIL.Image.AFFINE, (1, value, 0, 0, 1, 0)), value


def shear_y(img, value, max_value, bias=0):
    """Applies ShearY to an image."""
    value = _float_parameter(value, max_value) + bias
    if random.random() < 0.5:
        value = -value
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, value, 1, 0)), value


def solarize(img, value, max_value, bias=0):
    """Applies Solarize to an image."""
    value = _int_parameter(value, max_value) + bias
    return PIL.ImageOps.solarize(img, 256 - value), value


def translate_x(img, value, max_value, bias=0):
    """Applies TranslateX to an image."""
    value = _float_parameter(value, max_value) + bias
    if random.random() < 0.5:
        value = -value
    value = int(value * img.size[0])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, value, 0, 1, 0)), value


def translate_y(img, value, max_value, bias=0):
    """Applies TranslateX to an image."""
    value = _float_parameter(value, max_value) + bias
    if random.random() < 0.5:
        value = -value
    value = int(value * img.size[1])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, 0, 1, value)), value


def _float_parameter(value, max_value):
    return float(value) * max_value / PARAMETER_MAX


def _int_parameter(value, max_value):
    return int(value * max_value / PARAMETER_MAX)


rand_augment_pool = [
    (auto_contrast, None, None),
    (brightness, 0.9, 0.05),
    (color, 0.9, 0.05),
    (contrast, 0.9, 0.05),
    (equalize, None, None),
    (identity, None, None),
    (posterize, 4, 4),
    (rotate, 30, 0),
    (sharpness, 0.9, 0.05),
    (shear_x, 0.3, 0),
    (shear_y, 0.3, 0),
    (solarize, 256, 0),
    (translate_x, 0.3, 0),
    (translate_y, 0.3, 0),
]


# TODO: [Jihwan]: Can be removed by mmcls.datasets.pipeline.auto_augment Line 95 RandAugment class
@PIPELINES.register_module()
class OTXRandAugment:
    """RandAugment class for OTX classification."""

    def __init__(self, num_aug, magnitude, cutout_value=16):
        assert num_aug >= 1
        assert 1 <= magnitude <= 10
        self.num_aug = num_aug
        self.magnitude = magnitude
        self.cutout_value = cutout_value
        self.augment_pool = rand_augment_pool

    def __call__(self, results):
        """Call function of OTXRandAugment class."""
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            if not PIL.Image.isImageType(img):
                img = PIL.Image.fromarray(results[key])
            augs = random.choices(self.augment_pool, k=self.num_aug)
            for aug, max_value, bias in augs:
                value = np.random.randint(1, self.magnitude)
                if random.random() < 0.5:
                    img, value = aug(img, value=value, max_value=max_value, bias=bias)
                    results[f"rand_mc_{aug.__name__}"] = value
            img, xy, rec_color = cutout_abs(img, self.cutout_value)
            results["CutoutAbs"] = (xy, self.cutout_value, rec_color)
            results[key] = np.array(img)
        return results



"""Adapters for OTX Common Algorithm. - mmseg.model."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .backbones import LiteHRNet, MMOVBackbone
from .heads import CustomFCNHead, DetConHead, MMOVDecodeHead
from .losses import CrossEntropyLossWithIgnore, DetConLoss
from .necks import SelfSLMLP
from .schedulers import (
    ConstantScalarScheduler,
    PolyScalarScheduler,
    StepScalarScheduler,
)
from .segmentors import (
    DetConB,
    MeanTeacherSegmentor,
    SupConDetConB,
)

__all__ = [
    "LiteHRNet",
    "MMOVBackbone",
    "CustomFCNHead",
    "MMOVDecodeHead",
    "DetConLoss",
    "SelfSLMLP",
    "ConstantScalarScheduler",
    "PolyScalarScheduler",
    "StepScalarScheduler",
    "DetConB",
    "CrossEntropyLossWithIgnore",
    "SupConDetConB",
    "MeanTeacherSegmentor",
    "DetConHead",
]



#
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from datetime import datetime

import pytest

from otx.api.serialization.datetime_mapper import DatetimeMapper
from otx.api.utils.time_utils import now
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDatetimeMapper:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_serialization_deserialization(self):
        """
        This test serializes datetime, deserializes serialized datetime and compares with original one.
        """

        original_time = now()
        serialized_time = DatetimeMapper.forward(original_time)
        assert serialized_time == original_time.strftime("%Y-%m-%dT%H:%M:%S.%f")

        deserialized_time = DatetimeMapper.backward(serialized_time)
        assert original_time == deserialized_time

        deserialized_time = DatetimeMapper.backward(None)
        assert isinstance(deserialized_time, datetime)



import sys

import atheris
from helper import FuzzingHelper

from otx.cli.tools.cli import main as cli_main
from otx.cli.utils.errors import CliException


@atheris.instrument_func
def fuzz_otx(input_bytes):
    # create a FuzzingHelper instance to get suitable data type from the randomly generated 'input_bytes'
    helper = FuzzingHelper(input_bytes)
    backup_argv = sys.argv

    # get 'operation' arguments from 'input_bytes'
    operation = helper.get_string()
    sys.argv = ["otx", operation]
    try:
        _ = cli_main()
    except SystemExit as e:
        # argparser will throw SystemExit with code 2 when some required arguments are missing
        if e.code != 2:
            raise
    except CliException:
        pass
    # some known exceptions can be catched here
    finally:
        sys.argv = backup_argv


def main():
    # 'sys.argv' used to passing options to atheris.Setup()
    # available options can be found https://llvm.org/docs/LibFuzzer.html#options
    atheris.Setup(sys.argv, fuzz_otx)
    # Fuzz() will
    atheris.Fuzz()


if __name__ == "__main__":
    main()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import openvino.runtime as ov
import pytest
import torch

from otx.algorithms.detection.adapters.mmdet.models.dense_heads.mmov_rpn_head import (
    MMOVRPNHead,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMMOVSSDHead:
    @pytest.fixture(autouse=True)
    def setup(self):

        params = []
        results = []

        param = ov.opset10.parameter([1, 24, 64, 64], ov.Type.f32, name="in")
        filter = ov.opset10.constant(np.random.normal(size=(1, 24, 64, 64)), ov.Type.f32)
        mul = ov.opset10.matmul(param, filter, False, False)
        result_1 = ov.opset10.result(mul, name="cls_score_out")
        result_2 = ov.opset10.result(mul, name="bbox_pred_out")
        params.append(param)
        results.append(result_1)
        results.append(result_2)

        ov_model = ov.Model(results, params, "rpn_head")

        self.model = MMOVRPNHead(
            model_path_or_model=ov_model,
            transpose_reg=True,
            transpose_cls=True,
            inputs="in",
            outputs=[
                "cls_score_out",
                "bbox_pred_out",
            ],
            anchor_generator=dict(
                type="AnchorGenerator",
                base_sizes=[256],
                scales=[0.25, 0.5, 1, 2],
                ratios=[0.5, 1.0, 2.0],
                strides=[8],
            ),
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder",
                target_means=[0.0, 0.0, 0.0, 0.0],
                target_stds=[0.1, 0.1, 0.2, 0.2],
            ),
        )

    @e2e_pytest_unit
    def test_init_weights(self):
        self.model.init_weights()

    @e2e_pytest_unit
    def test_forward(self):
        data = {}
        for key, shape in self.model.model.input_shapes.items():
            shape = [1 if i == -1 else i for i in shape]
            data[key] = torch.randn(shape)
        self.model.train()
        self.model(list(data.values()))



"""Unit test for otx.algorithms.common.adapters.mmcv.hooks.model_ema_v2_hook."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.algorithms.common.adapters.mmcv.hooks.model_ema_v2_hook import (
    ModelEmaV2,
    ModelEmaV2Hook,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestModelEmaV2Hook:
    @e2e_pytest_unit
    def test_temp(self) -> None:
        try:
            hook = ModelEmaV2Hook()
            assert hook is None
        except Exception as e:
            print(e)
            pass


class TestModelEmaV2:
    @e2e_pytest_unit
    def test_temp(self) -> None:
        try:
            model = ModelEmaV2()
            assert model is None
        except Exception as e:
            print(e)
            pass



"""OPS (OperationRegistry) module for otx.core.ov.ops.builder."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Any, Optional

from otx.core.ov.registry import Registry


class OperationRegistry(Registry):
    """OperationRegistry class."""

    def __init__(self, name, add_name_as_attr=False):
        super().__init__(name, add_name_as_attr)
        self._registry_dict_by_type = {}

    def register(self, name: Optional[Any] = None):
        """Register function from name."""

        def wrap(obj):
            layer_name = name
            if layer_name is None:
                layer_name = obj.__name__
            layer_type = obj.TYPE
            layer_version = obj.VERSION
            assert layer_type and layer_version >= 0
            if self._add_name_as_attr:
                setattr(obj, self.REGISTERED_NAME_ATTR, layer_name)
            self._register(obj, layer_name, layer_type, layer_version)
            return obj

        return wrap

    def _register(self, obj, name, types, version):
        """Register function from obj and obj name."""
        super()._register(obj, name)
        if types not in self._registry_dict_by_type:
            self._registry_dict_by_type[types] = {}
        if version in self._registry_dict_by_type[types]:
            raise KeyError(f"{version} is already registered in {types}")
        self._registry_dict_by_type[types][version] = obj

    def get_by_name(self, name):
        """Get obj from name."""
        return self.get(name)

    def get_by_type_version(self, types, version):
        """Get obj from type and version."""
        if types not in self._registry_dict_by_type:
            raise KeyError(f"type {types} is not registered in {self._name}")
        if version not in self._registry_dict_by_type[types]:
            raise KeyError(f"version {version} is not registered in {types} of {self._name}")
        return self._registry_dict_by_type[types][version]


OPS = OperationRegistry("ov ops")



"""BaseTask for NNCF."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import io
import json
import os
from copy import deepcopy
from typing import Dict, List, Optional

import torch
from mmcv.utils import ConfigDict

import otx.algorithms.common.adapters.mmcv.nncf.patches  # noqa: F401  # pylint: disable=unused-import
from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    remove_from_config,
    remove_from_configs_by_type,
)
from otx.algorithms.common.adapters.nncf import (
    check_nncf_is_enabled,
    is_accuracy_aware_training_set,
)
from otx.algorithms.common.adapters.nncf.config import compose_nncf_config
from otx.algorithms.common.utils.callback import OptimizationProgressCallback
from otx.algorithms.common.utils.data import get_dataset
from otx.algorithms.common.utils.logger import get_logger
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import (
    OptimizationParameters,
    default_progress_callback,
)
from otx.api.entities.subset import Subset
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)

logger = get_logger()


class NNCFBaseTask(IOptimizationTask):  # pylint: disable=too-many-instance-attributes
    """NNCFBaseTask."""

    def __init__(self):
        check_nncf_is_enabled()
        self._nncf_data_to_build = None
        self._nncf_state_dict_to_build: Dict[str, torch.Tensor] = {}
        self._nncf_preset = None
        self._optimization_methods: List[OptimizationMethod] = []
        self._precision = [ModelPrecision.FP32]

        # Extra control variables.
        self._training_work_dir = None
        self._is_training = False
        self._should_stop = False
        self._optimization_type = ModelOptimizationType.NNCF
        self._time_monitor = None

        # Variables will be set in training backend task
        self._data_cfg = None
        self._model_ckpt = None
        self._model_dir = None
        self._labels = None
        self._recipe_cfg = None
        self._hyperparams = None
        self._task_environment = None

        logger.info("Task initialization completed")

    def _set_attributes_by_hyperparams(self):
        quantization = self._hyperparams.nncf_optimization.enable_quantization
        pruning = self._hyperparams.nncf_optimization.enable_pruning
        if quantization and pruning:
            self._nncf_preset = "nncf_quantization_pruning"
            self._optimization_methods = [
                OptimizationMethod.QUANTIZATION,
                OptimizationMethod.FILTER_PRUNING,
            ]
            self._precision = [ModelPrecision.INT8]
            return
        if quantization and not pruning:
            self._nncf_preset = "nncf_quantization"
            self._optimization_methods = [OptimizationMethod.QUANTIZATION]
            self._precision = [ModelPrecision.INT8]
            return
        if not quantization and pruning:
            self._nncf_preset = "nncf_pruning"
            self._optimization_methods = [OptimizationMethod.FILTER_PRUNING]
            self._precision = [ModelPrecision.FP32]
            return
        raise RuntimeError("Not selected optimization algorithm")

    def _init_train_data_cfg(self, dataset: DatasetEntity):
        logger.info("init data cfg.")
        data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val"],
            [Subset.TRAINING, Subset.VALIDATION],
        ):
            subset = get_dataset(dataset, subset)
            if subset:
                data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        return data_cfg

    def _init_nncf_cfg(self):
        nncf_config_path = os.path.join(self._model_dir, "compression_config.json")

        with open(nncf_config_path, encoding="UTF-8") as nncf_config_file:
            common_nncf_config = json.load(nncf_config_file)

        optimization_config = compose_nncf_config(common_nncf_config, [self._nncf_preset])

        max_acc_drop = self._hyperparams.nncf_optimization.maximal_accuracy_degradation / 100
        if "accuracy_aware_training" in optimization_config["nncf_config"]:
            # Update maximal_absolute_accuracy_degradation
            (
                optimization_config["nncf_config"]["accuracy_aware_training"]["params"][
                    "maximal_absolute_accuracy_degradation"
                ]
            ) = max_acc_drop
            # Force evaluation interval
            self._recipe_cfg.evaluation.interval = 1
        else:
            logger.info("NNCF config has no accuracy_aware_training parameters")

        return ConfigDict(optimization_config)

    def _prepare_optimize(self):
        assert self._recipe_cfg is not None

        # TODO: more delicate configuration change control in MPA side

        # last batch size of 1 causes undefined behaviour for batch normalization
        # when initializing and training NNCF
        if self._data_cfg is not None:
            data_loader = self._recipe_cfg.data.get("train_dataloader", ConfigDict())
            samples_per_gpu = data_loader.get("samples_per_gpu", self._recipe_cfg.data.get("samples_per_gpu"))
            otx_dataset = get_configs_by_keys(self._data_cfg.data.train, "otx_dataset")
            assert len(otx_dataset) == 1
            otx_dataset = otx_dataset[0]
            if otx_dataset is not None and len(otx_dataset) % samples_per_gpu == 1:
                data_loader["drop_last"] = True
                self._recipe_cfg.data["train_dataloader"] = data_loader

        # nncf does not suppoer FP16
        if "fp16" in self._recipe_cfg:
            remove_from_config(self._recipe_cfg, "fp16")
            logger.warning("fp16 option is not supported in NNCF. Switch to fp32.")

        # FIXME: nncf quantizer does not work with SAMoptimizer
        optimizer_config = self._recipe_cfg.optimizer_config
        if optimizer_config.get("type", "OptimizerHook") == "SAMOptimizerHook":
            optimizer_config.type = "OptimizerHook"
            logger.warning("Updateed SAMOptimizerHook to OptimizerHook as not supported.")

        # merge nncf_cfg
        nncf_cfg = self._init_nncf_cfg()
        self._recipe_cfg.merge_from_dict(nncf_cfg)

        # configure nncf
        nncf_config = self._recipe_cfg.get("nncf_config", {})
        if nncf_config.get("target_metric_name", None) is None:
            metric_name = self._recipe_cfg.evaluation.metric
            if isinstance(metric_name, list):
                metric_name = metric_name[0]
            nncf_config.target_metric_name = metric_name
            logger.info(f"'target_metric_name' not found in nncf config. Using {metric_name} as target metric")

        if is_accuracy_aware_training_set(nncf_config):
            # Prepare runner for Accuracy Aware
            self._recipe_cfg.runner = {
                "type": "AccuracyAwareRunner",
                "nncf_config": nncf_config,
            }

            # AccuracyAwareRunner needs to evaluate a model when it needs
            # unlike other runners counting on periodically evaluated score by 'EvalHook'.
            # To configure 'interval' to 'max_epoch' makes sure 'EvalHook' not to evaluate
            # during training.
            max_epoch = nncf_config.accuracy_aware_training.params.maximal_total_epochs
            self._recipe_cfg.evaluation.interval = max_epoch
            # Disable 'AdaptiveTrainSchedulingHook' as training is managed by AccuracyAwareRunner
            remove_from_configs_by_type(self._recipe_cfg.custom_hooks, "AdaptiveTrainSchedulingHook")

    @staticmethod
    def model_builder(
        config,
        *args,
        nncf_model_builder,
        model_config=None,
        data_config=None,
        is_export=False,
        return_compression_ctrl=False,
        **kwargs,
    ):
        """model_builder."""

        if model_config is not None or data_config is not None:
            config = deepcopy(config)
            if model_config is not None:
                config.merge_from_dict(model_config)
            if data_config is not None:
                config.merge_from_dict(data_config)

        compression_ctrl, model, = nncf_model_builder(
            config,
            distributed=False,
            *args,
            **kwargs,
        )

        if is_export:
            compression_ctrl.prepare_for_export()
            model.disable_dynamic_graph_building()

        if return_compression_ctrl:
            return compression_ctrl, model
        return model

    def _optimize(
        self,
        dataset: DatasetEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        raise NotImplementedError

    def _optimize_post_hook(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
    ):
        pass

    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        """NNCF Optimization."""
        if optimization_type is not OptimizationType.NNCF:
            raise RuntimeError("NNCF is the only supported optimization")

        if optimization_parameters is not None:
            update_progress_callback = optimization_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback

        self._time_monitor = OptimizationProgressCallback(
            update_progress_callback,
            loading_stage_progress_percentage=5,
            initialization_stage_progress_percentage=5,
        )

        self._data_cfg = self._init_train_data_cfg(dataset)
        self._is_training = True

        results = self._optimize(dataset, optimization_parameters)

        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            # output_model.model_status = ModelStatus.FAILED
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        self._optimize_post_hook(dataset, output_model)

        self.save_model(output_model)

        output_model.model_format = ModelFormat.BASE_FRAMEWORK
        output_model.optimization_type = self._optimization_type
        output_model.optimization_methods = self._optimization_methods
        output_model.precision = self._precision

        self._is_training = False

    def _save_model_post_hook(self, modelinfo):
        pass

    def save_model(self, output_model: ModelEntity):
        """Saving model function for NNCF Task."""
        assert self._recipe_cfg is not None

        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}

        model_ckpt = torch.load(self._model_ckpt, map_location=torch.device("cpu"))
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "VERSION": 1,
            "meta": {
                "nncf_enable_compression": True,
            },
        }
        self._save_model_post_hook(modelinfo)

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import cv2
import numpy as np
import pytest

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.exportable_code.visualizers import Visualizer
from otx.api.utils.shape_drawer import ShapeDrawer
from otx.api.utils.time_utils import now
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestVisualizer:
    image = np.random.randint(low=0, high=255, size=(480, 640, 3)).astype(np.float32)

    @staticmethod
    def annotation_scene() -> AnnotationSceneEntity:
        creation_date = now()
        annotation_color = Color(red=30, green=180, blue=70)
        other_annotation_color = Color(red=240, green=30, blue=40)
        detection_label = LabelEntity(
            name="detection label",
            domain=Domain.DETECTION,
            color=annotation_color,
            creation_date=creation_date,
            id=ID("detection_1"),
        )
        segmentation_label = LabelEntity(
            name="segmentation label",
            domain=Domain.SEGMENTATION,
            color=annotation_color,
            creation_date=creation_date,
            id=ID("segmentation_1"),
        )
        annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0.1, x2=0.4, y2=0.5),
            labels=[
                ScoredLabel(detection_label, 0.9),
                ScoredLabel(segmentation_label, 0.8),
            ],
        )

        classification_label = LabelEntity(
            name="classification label",
            domain=Domain.CLASSIFICATION,
            color=other_annotation_color,
            creation_date=creation_date,
            id=ID("classification_1"),
        )
        anomaly_segmentation_label = LabelEntity(
            name="anomaly_segmentation label",
            domain=Domain.ANOMALY_SEGMENTATION,
            color=other_annotation_color,
            creation_date=creation_date,
            id=ID("anomaly_segmentation_1"),
        )
        other_annotation = Annotation(
            shape=Ellipse(x1=0.6, y1=0.4, x2=0.7, y2=0.9),
            labels=[
                ScoredLabel(classification_label, 0.75),
                ScoredLabel(anomaly_segmentation_label, 0.9),
            ],
        )
        annotation_scene = AnnotationSceneEntity(
            annotations=[annotation, other_annotation],
            kind=AnnotationSceneKind.ANNOTATION,
        )
        return annotation_scene

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_visualizer_initialization(self):
        """
        <b>Description:</b>
        Check "Visualizer" class object initialization

        <b>Input data:</b>
        "Visualizer" object with specified attributes

        <b>Expected results:</b>
        Test passes if attributes of initialized "Visualizer" object are equal to expected

        <b>Steps</b>
        1. Check attributes of "Visualizer" object initialized with default optional parameters
        2. Check attributes of "Visualizer" object initialized with specified optional parameters
        """

        def check_visualizer_attributes(
            actual_visualizer: Visualizer,
            expected_name: str,
            expected_delay: int,
            expected_show_count: bool,
            expected_is_one_label: bool,
            expected_no_show: bool,
        ):
            assert actual_visualizer.window_name == expected_name
            assert actual_visualizer.delay == expected_delay
            assert actual_visualizer.no_show == expected_no_show
            assert isinstance(actual_visualizer.shape_drawer, ShapeDrawer)
            assert actual_visualizer.shape_drawer.show_count == expected_show_count
            assert actual_visualizer.shape_drawer.is_one_label == expected_is_one_label

        # Checking attributes of "Visualizer" initialized with default optional parameters
        visualizer = Visualizer()
        check_visualizer_attributes(
            actual_visualizer=visualizer,
            expected_name="Window",
            expected_delay=1,
            expected_show_count=False,
            expected_is_one_label=False,
            expected_no_show=False,
        )
        # Checking attributes of "Visualizer" initialized with specified optional parameters
        visualizer = Visualizer(
            window_name="Test Visualizer",
            show_count=True,
            is_one_label=True,
            no_show=True,
            delay=5,
        )
        check_visualizer_attributes(
            actual_visualizer=visualizer,
            expected_name="Test Visualizer",
            expected_delay=5,
            expected_show_count=True,
            expected_is_one_label=True,
            expected_no_show=True,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_visualizer_draw(self):
        """
        <b>Description:</b>
        Check "Visualizer" class "draw" method

        <b>Input data:</b>
        "Visualizer" object with specified attributes, "image" array, "annotation" AnnotationSceneEntity-type object

        <b>Expected results:</b>
        Test passes if array returned by "draw" method is equal to expected
        """
        annotation_scene = self.annotation_scene()
        image = self.image
        expected_image = image.copy()
        expected_image = cv2.cvtColor(expected_image, cv2.COLOR_RGB2BGR)
        shape_drawer = ShapeDrawer(show_count=False, is_one_label=False)
        expected_image = shape_drawer.draw(image=expected_image, entity=annotation_scene, labels=[])

        actual_image = Visualizer().draw(image=image, annotation=annotation_scene)
        assert np.array_equal(actual_image, expected_image)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_visualizer_no_show_mode(self):
        """
        <b>Description:</b>
        Check "Visualizer" class "no_show" parameter

        <b>Input data:</b>
        "Visualizer" object with specified attributes

        <b>Expected results:</b>
        Test passes if no exception is occured
        """
        self.annotation_scene()
        image = self.image
        visualizer = Visualizer(no_show=True)
        visualizer.show(image)
        visualizer.is_quit()



"""Utils for Action recognition OpenVINO export task."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from collections import OrderedDict
from typing import Any, Dict, Optional, Tuple

import numpy as np
import torch
from mmcv.runner import BaseModule
from mmcv.utils import Config
from mmdeploy.apis import build_task_processor
from mmdeploy.apis.core.pipeline_manager import no_mp
from mmdeploy.apis.onnx import export
from mmdeploy.backend.openvino.onnx2openvino import from_onnx
from mmdeploy.backend.openvino.utils import ModelOptimizerOptions

from otx.algorithms.action.adapters.mmaction.models import AVAFastRCNN


def _convert_sync_batch_to_normal_batch(module: BaseModule):
    """Convert the syncBNs into normal BN3ds."""
    module_output = module
    if isinstance(module, torch.nn.SyncBatchNorm):
        module_output = torch.nn.BatchNorm3d(
            module.num_features, module.eps, module.momentum, module.affine, module.track_running_stats
        )
        if module.affine:
            module_output.weight.data = module.weight.data.clone().detach()
            module_output.bias.data = module.bias.data.clone().detach()
            # keep requires_grad unchanged
            module_output.weight.requires_grad = module.weight.requires_grad
            module_output.bias.requires_grad = module.bias.requires_grad
        module_output.running_mean = module.running_mean
        module_output.running_var = module.running_var
        module_output.num_batches_tracked = module.num_batches_tracked
    for name, child in module.named_children():
        module_output.add_module(name, _convert_sync_batch_to_normal_batch(child))
    del module
    return module_output


# pylint: disable=too-many-instance-attributes
class Exporter:
    """Export class for action recognition model using mmdeploy framework."""

    def __init__(
        self, recipe_cfg: Config, weights: OrderedDict, deploy_cfg: Config, work_dir: str, half_precision: bool
    ):
        """Initialize Exporter.

        Args:
            recipe_cfg (Config): recipe config which contains model config
            weights (str): model weights
            deploy_cfg (Config): deploy config which contains deploy info
            work_dir (str): path to save onnx and openvino xml file
            half_precision (bool): whether to use half-precision(FP16)
        """

        self.task_processor = build_task_processor(recipe_cfg, deploy_cfg, "cpu")
        self.weights = weights

        self.deploy_cfg = deploy_cfg

        self.model = self._get_model()
        self.input_tensor, self.input_metas = self._get_inputs()
        self.work_dir = work_dir
        self.context_info = {"deploy_cfg": deploy_cfg}
        if half_precision:
            self.deploy_cfg.backend_config.mo_options["flags"] = ["--compress_to_fp16"]

    def _get_model(self) -> torch.nn.Module:
        """Prepare torch model for exporting."""

        model = self.task_processor.init_pytorch_model(None)
        model.load_state_dict(self.weights)
        if isinstance(model, AVAFastRCNN):
            model.patch_for_export()
        return _convert_sync_batch_to_normal_batch(model)

    def _get_inputs(self) -> Tuple[torch.Tensor, Optional[Dict[str, Any]]]:
        """Prepare torch model's input and input_metas."""

        height, width = self.deploy_cfg.backend_config.model_inputs[0]["opt_shapes"]["input"][-2:]
        if isinstance(self.model, AVAFastRCNN):
            input_tensor = torch.randn(1, 3, 32, height, width)
            input_metas = {
                "img_metas": [
                    [
                        {
                            "ori_shape": (height, width),
                            "img_shape": (height, width),
                            "pad_shape": (height, width),
                            "scale_factor": np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float32),
                        }
                    ]
                ]
            }
        else:
            input_tensor = torch.randn(1, 1, 3, 32, height, width)
            input_metas = None
        return input_tensor, input_metas

    def export(self):
        """Export action model using mmdeploy apis."""

        with no_mp():
            export(
                self.model,
                self.input_tensor,
                self.work_dir,
                "onnxruntime",
                self.input_metas,
                self.context_info,
                self.deploy_cfg.ir_config.input_names,
                self.deploy_cfg.ir_config.output_names,
            )

            from_onnx(
                self.work_dir + ".onnx",
                self.work_dir.replace("openvino", ""),
                {self.deploy_cfg.ir_config.input_names[0]: self.input_tensor.shape},
                self.deploy_cfg.ir_config.output_names,
                ModelOptimizerOptions(self.deploy_cfg.backend_config.mo_options),
            )



"""Module for defining AsymmetricLossWithIgnore."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcls.models.builder import LOSSES
from mmcls.models.losses.utils import weight_reduce_loss
from torch import nn


def asymmetric_loss_with_ignore(
    pred,
    target,
    valid_label_mask=None,
    weight=None,
    gamma_pos=1.0,
    gamma_neg=4.0,
    clip=0.05,
    reduction="none",
    avg_factor=None,
):  # pylint: disable=too-many-arguments
    """Asymmetric loss, please refer to the `paper <https://arxiv.org/abs/2009.14119>`_ for details.

    Args:
        pred (torch.Tensor): The prediction with shape (N, *).
        target (torch.Tensor): The ground truth label of the prediction with
            shape (N, *).
        valid_label_mask (torch.Tensor, optional): Label mask for consideration of ignored label.
        weight (torch.Tensor, optional): Sample-wise loss weight with shape
            (N, ). Dafaults to None.
        gamma_pos (float): positive focusing parameter. Defaults to 0.0.
        gamma_neg (float): Negative focusing parameter. We usually set
            gamma_neg > gamma_pos. Defaults to 4.0.
        clip (float, optional): Probability margin. Defaults to 0.05.
        reduction (str): The method used to reduce the loss.
            Options are "none", "mean" and "sum". If reduction is 'none' , loss
             is same shape as pred and label. Defaults to 'mean'.
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.

    Returns:
        torch.Tensor: Loss.
    """
    assert pred.shape == target.shape, "pred and target should be in the same shape."

    eps = 1e-8
    pred_sigmoid = pred.sigmoid()
    target = target.type_as(pred)
    if reduction != "mean":  # we don't use avg factor with other reductions
        avg_factor = None  # if we are not set this to None the exception will be throwed

    if clip and clip > 0:
        pos_target = (1 - pred_sigmoid + clip).clamp(max=1) * (1 - target) + pred_sigmoid * target
    else:
        pos_target = (1 - pred_sigmoid) * (1 - target) + pred_sigmoid * target
    asymmetric_weight = (1 - pos_target).pow(gamma_pos * target + gamma_neg * (1 - target))
    loss = -torch.log(pos_target.clamp(min=eps)) * asymmetric_weight

    if valid_label_mask is not None:
        loss = loss * valid_label_mask

    if weight is not None:
        assert weight.dim() == 1
        weight = weight.float()
        if pred.dim() > 1:
            weight = weight.reshape(-1, 1)
    loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
    return loss


@LOSSES.register_module()
class AsymmetricLossWithIgnore(nn.Module):
    """Asymmetric loss.

    Args:
        gamma_pos (float): positive focusing parameter.
            Defaults to 0.0.
        gamma_neg (float): Negative focusing parameter. We
            usually set gamma_neg > gamma_pos. Defaults to 4.0.
        clip (float, optional): Probability margin. Defaults to 0.05.
        reduction (str): The method used to reduce the loss into
            a scalar.
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(self, gamma_pos=0.0, gamma_neg=4.0, clip=0.05, reduction="none", loss_weight=1.0):
        super().__init__()
        self.gamma_pos = gamma_pos
        self.gamma_neg = gamma_neg
        self.clip = clip
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self, pred, target, valid_label_mask=None, weight=None, avg_factor=None, reduction_override=None):
        """Forward fuction of asymmetric loss."""
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction
        loss_cls = self.loss_weight * asymmetric_loss_with_ignore(
            pred,
            target,
            valid_label_mask,
            weight,
            gamma_pos=self.gamma_pos,
            gamma_neg=self.gamma_neg,
            clip=self.clip,
            reduction=reduction,
            avg_factor=avg_factor,
        )
        return loss_cls



import os
import socket
from contextlib import closing
from copy import deepcopy

import pytest
import torch

from otx.cli.utils import multi_gpu
from otx.cli.utils.multi_gpu import (
    MultiGPUManager,
    _get_free_port,
    get_gpu_ids,
    is_multigpu_child_process,
    set_arguments_to_argv,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit

NUM_AVAILABLE_GPU = torch.cuda.device_count()


@e2e_pytest_unit
def test_get_free_port():
    free_port = _get_free_port()

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(("", free_port))


@e2e_pytest_unit
def test_get_gpu_ids():
    gpus = []
    for i in range(0, NUM_AVAILABLE_GPU, 2):
        gpus.append(i)

    expected_result = deepcopy(gpus)
    gpus.append(NUM_AVAILABLE_GPU + 2)

    assert get_gpu_ids(",".join([str(val) for val in gpus])) == expected_result


@e2e_pytest_unit
def test_get_gpu_ids_with_wrong_args():
    with pytest.raises(ValueError):
        get_gpu_ids("abcd")


@pytest.fixture
def mock_argv_without_params(mocker):
    mock_sys = mocker.patch("otx.cli.utils.multi_gpu.sys")
    mock_sys.argv = ["--a_key", "a_val", "--b_key"]
    return mock_sys.argv


@pytest.fixture
def mock_argv_with_params(mock_argv_without_params):
    mock_argv_without_params.extend(["params", "--c_key", "c_val", "--d_key"])
    return mock_argv_without_params


@e2e_pytest_unit
def test_set_arguments_to_argv_key_exist(mock_argv_without_params):
    """Test a case where key already exists and value exists."""
    other_val = "other_val"
    set_arguments_to_argv("--a_key", other_val)

    assert mock_argv_without_params[1] == other_val


@e2e_pytest_unit
def test_set_arguments_to_argv_keys_exist(mock_argv_without_params):
    """Test a case where key already exists and value exists."""
    other_val = "other_val"
    set_arguments_to_argv(["--a_key", "-a"], other_val)

    assert mock_argv_without_params[1] == other_val


@e2e_pytest_unit
def test_set_arguments_to_argv_key_exist_none_val(mock_argv_without_params):
    """Test a case where key already exists in argv and value doesn't exists."""
    expected_result = deepcopy(mock_argv_without_params)
    set_arguments_to_argv("--a_key")

    assert mock_argv_without_params == expected_result


@e2e_pytest_unit
def test_set_arguments_to_argv_key(mock_argv_with_params):
    """Test a case where key to set doesn't exists in argv and order of key is before params and vlaue exists."""
    set_arguments_to_argv("--other_key", "other_val")

    param_idx = mock_argv_with_params.index("params")
    new_key_idx = mock_argv_with_params.index("--other_key")

    assert new_key_idx < param_idx
    assert mock_argv_with_params[new_key_idx + 1] == "other_val"


@e2e_pytest_unit
def test_set_arguments_to_argv_key_none_val(mock_argv_with_params):
    """Test a case where key to set doesn't exists in argv and order of key is before params and vlaue doesn't exist."""
    set_arguments_to_argv("--other_key")

    param_idx = mock_argv_with_params.index("params")
    new_key_idx = mock_argv_with_params.index("--other_key")

    assert new_key_idx < param_idx
    assert "--other_key" in mock_argv_with_params


@e2e_pytest_unit
def test_set_arguments_to_argv_key_after_param(mock_argv_with_params):
    """Test a case where key to set doesn't exists in argv and order of key is after params and vlaue exists."""
    set_arguments_to_argv("--other_key", "other_val", True)

    param_idx = mock_argv_with_params.index("params")
    new_key_idx = mock_argv_with_params.index("--other_key")

    assert new_key_idx > param_idx
    assert mock_argv_with_params[new_key_idx + 1] == "other_val"


@e2e_pytest_unit
def test_set_arguments_to_argv_key_after_param_non_val(mock_argv_with_params):
    """Test a case where key to set doesn't exists in argv and order of key is after params and vlaue doesn't exist."""
    set_arguments_to_argv("--other_key", after_params=True)

    param_idx = mock_argv_with_params.index("params")
    new_key_idx = mock_argv_with_params.index("--other_key")

    assert new_key_idx > param_idx
    assert "--other_key" in mock_argv_with_params


@e2e_pytest_unit
def test_is_multigpu_child_process(mocker):
    mocker.patch.object(multi_gpu.dist, "is_initialized", return_value=True)
    os.environ["LOCAL_RANK"] = "1"
    assert is_multigpu_child_process()


@e2e_pytest_unit
def test_is_multigpu_child_process_no_initialized(mocker):
    mocker.patch.object(multi_gpu.dist, "is_initialized", return_value=False)
    os.environ["LOCAL_RANK"] = "1"
    assert not is_multigpu_child_process()


@e2e_pytest_unit
def test_is_multigpu_child_process_rank0(mocker):
    mocker.patch.object(multi_gpu.dist, "is_initialized", return_value=True)
    os.environ["LOCAL_RANK"] = "0"
    assert not is_multigpu_child_process()


class TestMultiGPUManager:
    @pytest.fixture(autouse=True)
    def _set_up(self, mocker):
        self.mock_singal = mocker.patch.object(multi_gpu, "signal")
        self.mock_thread = mocker.patch.object(multi_gpu.threading, "Thread")
        self.mock_train_func = mocker.MagicMock()
        self.mock_mp = mocker.patch.object(multi_gpu, "mp")
        self.mock_process = mocker.MagicMock()
        self.mock_mp.get_context.return_value.Process = self.mock_process
        self.num_gpu = 4

        mock_torch = mocker.patch.object(multi_gpu, "torch")
        mock_torch.cuda.device_count.return_value = self.num_gpu

        self.multigpu_manager = MultiGPUManager(self.mock_train_func, ",".join([str(i) for i in range(self.num_gpu)]))

    @pytest.fixture
    def process_arr(self, mocker):
        """List consists of normal process excpet last one. Last element is a process which exit abnormally."""
        normal_process = mocker.MagicMock()
        normal_process.is_alive.return_value = True
        wrong_process = mocker.MagicMock()
        wrong_process.is_alive.return_value = False
        wrong_process.exitcode = 1
        process_arr = []
        for _ in range(self.num_gpu - 2):
            process_arr.append(deepcopy(normal_process))
        process_arr.append(wrong_process)

        return process_arr

    @e2e_pytest_unit
    def test_init(self, mocker):
        MultiGPUManager(mocker.MagicMock(), "0,1", "localhost:0")

    @e2e_pytest_unit
    @pytest.mark.parametrize("num_gpu", [4, 10])
    def test_is_available(self, mocker, num_gpu):
        multigpu_manager = MultiGPUManager(
            mocker.MagicMock(), ",".join([str(i) for i in range(num_gpu)]), "localhost:0"
        )

        assert multigpu_manager.is_available()

    @e2e_pytest_unit
    def test_is_unavailable(self, mocker):
        mock_torch = mocker.patch.object(multi_gpu, "torch")
        mock_torch.cuda.device_count.return_value = 0
        multigpu_manager = MultiGPUManager(mocker.MagicMock(), ",".join([str(i) for i in range(4)]), "localhost:0")

        assert not multigpu_manager.is_available()

    @e2e_pytest_unit
    def test_is_unavailable_by_torchrun(self, mocker):
        mock_os = mocker.patch.object(multi_gpu, "os")
        mock_os.environ = {"TORCHELASTIC_RUN_ID": "1234"}
        multigpu_manager = MultiGPUManager(mocker.MagicMock(), ",".join([str(i) for i in range(4)]), "localhost:0")

        assert not multigpu_manager.is_available()

    @e2e_pytest_unit
    def test_setup_multi_gpu_train(self, mocker):
        # prepare
        mock_initialize_multigpu_train = mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        mock_hyper_parameters = mocker.MagicMock()
        mock_hyper_parameters.learning_parameters.learning_rate = 0.01
        mock_hyper_parameters.learning_parameters.batch_size = 8
        mock_sys = mocker.patch.object(multi_gpu, "sys")
        mock_sys.argv = []
        fake_output_path = "fake"

        # run
        self.multigpu_manager.setup_multi_gpu_train(fake_output_path, mock_hyper_parameters)

        # check spwaning child process
        assert self.mock_process.call_count == self.num_gpu - 1
        assert self.mock_process.return_value.start.call_count == self.num_gpu - 1
        assert self.mock_process.call_args.kwargs["target"] == MultiGPUManager.run_child_process
        assert self.mock_process.call_args.kwargs["args"][0] == self.mock_train_func  # train_func
        assert self.mock_process.call_args.kwargs["args"][1] == fake_output_path  # output_path
        assert self.mock_process.call_args.kwargs["args"][-1] == self.num_gpu  # num_gpu

        # check initialize multigpu trian
        mock_initialize_multigpu_train.assert_called_once()
        assert mock_initialize_multigpu_train.call_args.args[-1] == self.num_gpu  # world_size
        assert mock_initialize_multigpu_train.call_args.args[-2] == list(range(self.num_gpu))  # gpu_ids

        # check that making a thread to check child process is alive
        self.mock_thread.assert_called_once_with(target=self.multigpu_manager._check_child_processes_alive, daemon=True)
        self.mock_thread.return_value.start.assert_called_once()

        # check that register signal callback
        assert self.mock_singal.signal.call_count == 2
        mock_singal_args = self.mock_singal.signal.call_args_list
        assert mock_singal_args[0][0][0] in (self.mock_singal.SIGINT, self.mock_singal.SIGTERM)
        assert mock_singal_args[1][0][0] in (self.mock_singal.SIGINT, self.mock_singal.SIGTERM)
        assert mock_singal_args[0][0][1] == self.multigpu_manager._terminate_signal_handler
        assert mock_singal_args[1][0][1] == self.multigpu_manager._terminate_signal_handler

        # check that optimized hyper parameters are in sys.argv to pass them to child process
        assert "--learning_parameters.learning_rate" in mock_sys.argv
        assert mock_sys.argv[mock_sys.argv.index("--learning_parameters.learning_rate") + 1] == "0.01"
        assert "--learning_parameters.batch_size" in mock_sys.argv
        assert mock_sys.argv[mock_sys.argv.index("--learning_parameters.batch_size") + 1] == "8"

    @e2e_pytest_unit
    def test_check_child_processes_alive(self, mocker, process_arr):
        # prepare
        mock_kill = mocker.patch.object(multi_gpu.os, "kill")
        mocker.patch.object(multi_gpu.time, "sleep")
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        self.mock_process.side_effect = process_arr

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        self.multigpu_manager._check_child_processes_alive()

        # check
        for p in process_arr[: self.num_gpu - 2]:
            p.kill.assert_called_once()
        mock_kill.assert_called_once_with(os.getpid(), self.mock_singal.SIGKILL)

    @e2e_pytest_unit
    def test_terminate_signal_handler(self, mocker, process_arr):
        # prepare
        mock_exit = mocker.patch.object(multi_gpu.sys, "exit")
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        self.mock_process.side_effect = process_arr

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        self.multigpu_manager._terminate_signal_handler(2, mocker.MagicMock())

        # check
        for p in process_arr[: self.num_gpu - 2]:
            p.kill.assert_called_once()
        mock_exit.assert_called_once()

    @e2e_pytest_unit
    def test_terminate_signal_handler_not_main_thread(self, mocker, process_arr):
        # prepare
        def raise_error(*args, **kwargs):
            raise RuntimeError

        mock_exit = mocker.patch.object(multi_gpu.sys, "exit")
        mock_exit.side_effect = raise_error
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        mocker.patch.object(multi_gpu.os, "getpid").return_value = os.getpid() + 1
        self.mock_process.side_effect = process_arr

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        with pytest.raises(RuntimeError):
            self.multigpu_manager._terminate_signal_handler(2, mocker.MagicMock())

        # check
        for p in process_arr[: self.num_gpu - 2]:
            p.kill.assert_not_called()

    @e2e_pytest_unit
    def test_finalize(self, mocker, process_arr):
        # prepare
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        self.mock_process.side_effect = process_arr

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        self.multigpu_manager.finalize()

        # check
        for p in process_arr:
            p.join.assert_called_once()

    @e2e_pytest_unit
    def test_finalize_still_running_child_process(self, mocker, process_arr):
        # prepare
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        self.mock_process.side_effect = process_arr
        for p in process_arr:
            p.exitcode = None
            p.join.return_value = None

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        self.multigpu_manager.finalize()

        # check
        for p in process_arr:
            p.join.assert_called_once()
            p.kill.assert_called_once()

    @e2e_pytest_unit
    def test_finalize_before_spawn(self, mocker, process_arr):
        # prepare
        mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        self.mock_process.side_effect = process_arr

        # run
        self.multigpu_manager.setup_multi_gpu_train("fake")
        self.multigpu_manager.finalize()

    @e2e_pytest_unit
    def test_initialize_multigpu_train(self, mocker):
        # prepare
        mock_os = mocker.patch.object(multi_gpu, "os")
        mock_os.environ = {}
        mocker.patch.object(multi_gpu.dist, "get_world_size", return_value=2)
        mocker.patch.object(multi_gpu.dist, "get_rank", return_value=0)

        # run
        MultiGPUManager.initialize_multigpu_train(
            rdzv_endpoint="localhost:1234",
            rank=0,
            local_rank=0,
            gpu_ids=[0, 1],
            world_size=2,
        )

        # check
        assert mock_os.environ["MASTER_ADDR"] == "localhost"
        assert mock_os.environ["MASTER_PORT"] == "1234"
        assert mock_os.environ["LOCAL_WORLD_SIZE"] == "2"
        assert mock_os.environ["WORLD_SIZE"] == "2"
        assert mock_os.environ["LOCAL_RANK"] == "0"
        assert mock_os.environ["RANK"] == "0"

    @e2e_pytest_unit
    def test_run_child_process(self, mocker):
        # prepare
        mock_set_start_method = mocker.patch.object(multi_gpu.mp, "set_start_method")
        mock_sys = mocker.patch.object(multi_gpu, "sys")
        mock_sys.argv = ["--gpus", "0,1"]
        output_path = "mock_output_path"
        rdzv_endpoint = "localhost:1234"
        mock_initialize_multigpu_train = mocker.patch.object(MultiGPUManager, "initialize_multigpu_train")
        mock_threading = mocker.patch.object(multi_gpu, "threading")
        mock_train_func = mocker.MagicMock()

        # run
        MultiGPUManager.run_child_process(
            train_func=mock_train_func,
            output_path=output_path,
            rdzv_endpoint=rdzv_endpoint,
            rank=0,
            local_rank=0,
            gpu_ids=[0, 1],
            world_size=4,
        )

        # check
        assert mock_set_start_method.call_args.kwargs["method"] is None
        assert "--gpus" not in mock_sys.argv
        for output_arg_key in ["-o", "--output", False]:
            if output_arg_key in mock_sys.argv:
                break
        assert output_arg_key is not False, "There arn't both '-o' and '--output'."
        assert mock_sys.argv[mock_sys.argv.index(output_arg_key) + 1] == output_path
        assert "--rdzv-endpoint" in mock_sys.argv
        assert mock_sys.argv[mock_sys.argv.index("--rdzv-endpoint") + 1] == rdzv_endpoint
        mock_initialize_multigpu_train.assert_called_once()
        mock_threading.Thread.assert_called_once_with(target=MultiGPUManager.check_parent_processes_alive, daemon=True)
        mock_threading.Thread.call_args.return_value.start.assert_called_once
        mock_train_func.assert_called_once()

    @e2e_pytest_unit
    def test_check_parent_processes_alive(self, mocker):
        # prepare
        mocker.patch.object(multi_gpu, "time")
        mock_cur_process = mocker.MagicMock()
        mocker.patch.object(multi_gpu.psutil, "Process", return_value=mock_cur_process)
        mock_cur_process.parent.return_value.is_running.return_value = False

        # run
        MultiGPUManager.check_parent_processes_alive()

        # check
        mock_cur_process.kill.assert_called_once()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch

from otx.core.ov.ops.arithmetics import AddV1, DivideV1, MultiplyV1, SubtractV1, TanV0
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMultiplyV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)

    @e2e_pytest_unit
    def test_forward(self):
        op = MultiplyV1("dummy", shape=self.shape, auto_broadcast="none")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 * input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        with pytest.raises(AssertionError):
            output = op(input_1, input_2)

        op = MultiplyV1("dummy", shape=self.shape, auto_broadcast="numpy")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 * input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 * input_2)

        op = MultiplyV1("dummy", shape=self.shape, auto_broadcast="dummy")
        with pytest.raises(NotImplementedError):
            output = op(input_1, input_2)


class TestDivideV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)

    @e2e_pytest_unit
    def test_forward(self):
        op = DivideV1("dummy", shape=self.shape, auto_broadcast="none")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 / input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        with pytest.raises(AssertionError):
            output = op(input_1, input_2)

        op = DivideV1("dummy", shape=self.shape, auto_broadcast="numpy")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 / input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 / input_2)

        op = DivideV1("dummy", shape=self.shape, auto_broadcast="dummy")
        with pytest.raises(NotImplementedError):
            output = op(input_1, input_2)


class TestAddV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)

    @e2e_pytest_unit
    def test_forward(self):
        op = AddV1("dummy", shape=self.shape, auto_broadcast="none")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 + input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        with pytest.raises(AssertionError):
            output = op(input_1, input_2)

        op = AddV1("dummy", shape=self.shape, auto_broadcast="numpy")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 + input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 + input_2)

        op = AddV1("dummy", shape=self.shape, auto_broadcast="dummy")
        with pytest.raises(NotImplementedError):
            output = op(input_1, input_2)


class TestSubtractV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)

    @e2e_pytest_unit
    def test_forward(self):
        op = SubtractV1("dummy", shape=self.shape, auto_broadcast="none")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 - input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        with pytest.raises(AssertionError):
            output = op(input_1, input_2)

        op = SubtractV1("dummy", shape=self.shape, auto_broadcast="numpy")
        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape)
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 - input_2)

        input_1 = torch.randn(self.shape)
        input_2 = torch.randn(self.shape[1:])
        output = op(input_1, input_2)
        assert torch.equal(output, input_1 - input_2)

        op = SubtractV1("dummy", shape=self.shape, auto_broadcast="dummy")
        with pytest.raises(NotImplementedError):
            output = op(input_1, input_2)


class TestTanV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = TanV0("dummy", shape=self.shape)
        output = op(self.input)
        assert torch.equal(output, torch.tan(self.input))
